/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"





// things do do after a basic timestep
void
AMRNavierStokes::postTimeStep()
{

  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::postTimeStep " << m_level << endl;
  } 

  // none of these things are done if this is the finest level
  
  if (!finestLevel()) {

    AMRNavierStokes* fineAMRNavierStokesPtr = finerNSPtr();
    // create boundary condition object
    bool isViscous = (s_nu > 0.0);
    basicVelBC* velBCPtr = m_physBCPtr->uStarBC(isViscous);
    

    // first do refluxing and avgDown for conservation
    // do momentum refluxing first
    Real refluxScale = 1.0/m_dx;
    LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
    if (s_reflux_momentum) {
      if (s_implicit_reflux) {
	// defer this until we're doing sync projection
      } else {
	m_flux_register.reflux(newVel, refluxScale);
	CoarseAverage& velAvgDown = fineAMRNavierStokesPtr->m_coarse_average;
	velAvgDown.averageToCoarse(newVel,*fineAMRNavierStokesPtr->m_vel_new_ptr);
      }
    }

    // now do lamdba
    LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
    m_lambda_flux_reg.reflux(newLambda,refluxScale);
    CoarseAverage& lambdaAvgDown 
      = fineAMRNavierStokesPtr->m_coarse_average_lambda;
    lambdaAvgDown.averageToCoarse(newLambda,*fineAMRNavierStokesPtr->m_lambda_new_ptr);

    // finally, do scalars -- loop over components
    if (s_reflux_scal) {
      if (s_implicit_scal_reflux) {
	// if implicit refluxing, defer until we're doing multilevel stuff
      } else { 
	for (int comp=0; comp<s_num_scal_comps; comp++) {
	  LevelData<FArrayBox>& newScalars = *m_scal_new[comp];
	  m_scal_fluxreg_ptrs[comp]->reflux(newScalars, refluxScale);
	  CoarseAverage& scalAvgDown = 
	    fineAMRNavierStokesPtr->m_coarse_average_lambda;
	  
	  scalAvgDown.averageToCoarse(newScalars,
				      *fineAMRNavierStokesPtr->m_scal_new[comp]);
	}
      }
    } // end if reflux scalars

    // now call sync projection if necessary
    Real crseTime = 0.0;
    if (m_level > 0) crseTime = m_coarser_level_ptr->time();


    // do multilevel operations if this is the coarsest level or if
    // coarser level is not at the same time as this level
    if (m_level == 0 || (abs(crseTime-m_time)>TIME_EPS)) {

      int finest_level = m_level;
      // index through levels to find out what finest level is
      // (there must be a better way to do this!)
      AMRNavierStokes* thisNSPtr = this;
      while (!thisNSPtr->finestLevel()) {
	thisNSPtr = thisNSPtr->finerNSPtr();
      }
      assert(thisNSPtr->finestLevel());
      finest_level = thisNSPtr->m_level;

      // first, do implicit refluxing for scalars
      if (s_implicit_scal_reflux && s_reflux_scal) {
	Vector<LevelData<FArrayBox>* > scalRefluxCorr(finest_level+1,NULL);
	Vector<LevelData<FArrayBox>* > scalRefluxRHS(finest_level+1,NULL);
	Vector<DisjointBoxLayout> AMRgrids(finest_level+1);
	Vector<int> amrRefRatios(finest_level+1);
	Vector<Real> AmrDx(finest_level+1);
	Vector<ProblemDomain> AmrDomains(finest_level+1);

	// loop over levels, allocate storage, set up for AMRSolve
	thisNSPtr = this;
	int startLev=m_level;
	// if crser level exists, define it as well for BC's
	if (startLev > 0) {
	  startLev = startLev-1;
	  thisNSPtr = thisNSPtr->crseNSPtr();
	}
	for (int lev=startLev; lev<=finest_level; lev++) {
	  const DisjointBoxLayout& levelGrids = thisNSPtr->newVel().getBoxes();
	  // recall that AMRSolver can only do one component at a time.
	  // rhs has no ghost cells
	  scalRefluxRHS[lev] = new LevelData<FArrayBox>(levelGrids, 1);
	  //soln has one layer of ghost cells
	  IntVect ghostVect(D_DECL(1,1,1));
	  scalRefluxCorr[lev] = new LevelData<FArrayBox>(levelGrids,
							 1,ghostVect);
	  AMRgrids[lev] = levelGrids;
	  amrRefRatios[lev] = thisNSPtr->refRatio();
	  AmrDx[lev] = thisNSPtr->Dx();
	  AmrDomains[lev] = thisNSPtr->problemDomain();

	  // initialize corr, RHS to 0 
	  DataIterator levelDit = scalRefluxRHS[lev]->dataIterator();
	  LevelData<FArrayBox>& levelRefluxRHS = *(scalRefluxRHS[lev]);
	  LevelData<FArrayBox>& levelRefluxCorr = *(scalRefluxCorr[lev]);
	  for (levelDit.reset(); levelDit.ok(); ++levelDit) {
	    levelRefluxRHS[levelDit()].setVal(0.0);
	    levelRefluxCorr[levelDit()].setVal(0.0);
	  }
	  thisNSPtr = thisNSPtr->finerNSPtr();
	}
	
	

	Interval solverComps(0,0);

	// now do each scalar component
	for (int scalComp=0; scalComp<s_num_scal_comps; scalComp++) {

	  DomainGhostBC* scalPhysBCPtr = m_physBCPtr->scalarRefluxSolveBC(scalComp);

	  // loop over levels and establish RHS, corr
	  thisNSPtr = this;
	  
	  for (int lev=m_level; lev<=finest_level; lev++) {
	    
	    if (lev < finest_level) {
	      LevelData<FArrayBox>& levelRefluxRHS = *(scalRefluxRHS[lev]);
	      // first set to 0
	      DataIterator levelDit = levelRefluxRHS.dataIterator();
	      for (levelDit.reset(); levelDit.ok(); ++levelDit) {
		levelRefluxRHS[levelDit()].setVal(0.0);
	      }
	      refluxScale = 1.0/AmrDx[lev];
	      thisNSPtr->m_scal_fluxreg_ptrs[scalComp]->reflux(levelRefluxRHS, 
							       refluxScale);
	      // initial guess for correction is RHS
	      levelRefluxRHS.copyTo(solverComps, *(scalRefluxCorr[lev]),
				    solverComps);
	    }
	    
	    thisNSPtr = thisNSPtr->finerNSPtr();
	  } // end loop over levels to compute RHS

	  // define levelOp and solver for diffusive solves; this is 
	  // redundant and wasteful, but need to define
	  // new AMRSolver for each component because they 
	  // may have different coefficients

	  Real nuComp = s_scal_coeffs[scalComp];
	  // only do all of this if the darn thing is actually diffusive
	  if (nuComp > 0) {
	    HelmholtzOp diffusiveOp;
	    diffusiveOp.setDomainGhostBC(*scalPhysBCPtr);
	    Real beta = -nuComp*m_dt;
	    diffusiveOp.setHelmCoeff(beta);
	    
	    // don't need this anymore...
	    delete scalPhysBCPtr;
	  
	    // now define multilevel solver for diffusive solves
	    AMRSolver diffusionSolver(AMRgrids, AmrDomains,
				      AmrDx, amrRefRatios, finest_level+1,
				      m_level, &diffusiveOp);

	    diffusionSolver.setTolerance(s_viscous_solver_tol);
	    if (s_verbosity >= 3) {
	      diffusionSolver.setVerbose(true);
	    } else {
	      diffusionSolver.setVerbose(false);
	    }

	    // quick check -- want sum(RHS) to be 0
#ifdef DEBUG
	    Interval solverComps(0,0); 
	    Real sumRHS;
	    sumRHS = computeSum(scalRefluxRHS, amrRefRatios, m_dx,
				solverComps, m_level);
	    pout() << "Sum(RHS) for implicit reflux for " 
                   << s_scal_names[scalComp] << " = " 
                   << setiosflags(ios::scientific) << sumRHS << endl;
#endif

	    // now solve
	    diffusionSolver.solveAMR(scalRefluxCorr, scalRefluxRHS);

	    // now increment scalars with reflux correction
	    // go from finest->coarsest so that we can also avgDown
	    // increment NS pointer to finest level
	    thisNSPtr = this;
	    while (thisNSPtr->finerNSPtr() != NULL) {
	      thisNSPtr = thisNSPtr->finerNSPtr();
	    }
	    assert(thisNSPtr->m_level == finest_level);

	    for (int lev=finest_level; lev>=m_level; lev--) {
	      LevelData<FArrayBox>& levelScal = thisNSPtr->newScal(scalComp);
	      LevelData<FArrayBox>& levelCorr = *(scalRefluxCorr[lev]);
	      DataIterator levelDit = levelCorr.dataIterator();
	      for (levelDit.reset(); levelDit.ok(); ++levelDit) {
		levelScal[levelDit()] += levelCorr[levelDit()];
	      }

	      // if a finer level exists, do avgDown as well
	      if (lev < finest_level) {
		AMRNavierStokes& fineNS = *(thisNSPtr->finerNSPtr());
		// quick sanity check
		assert (fineNS.m_level = lev+1);
		CoarseAverage& scalAvgDown = fineNS.m_coarse_average_lambda;
		LevelData<FArrayBox>& fineScal = fineNS.newScal(scalComp);
		scalAvgDown.averageToCoarse(newScal(scalComp), fineScal);
	      }
	      thisNSPtr = thisNSPtr->crseNSPtr();
	    } // end loop over levels

	  } else {
	    // if this component is not diffusive, just add RHS to scalar
	    // now increment scalars with reflux correction
	    // go from finest->coarsest so that we can also avgDown
	    // increment NS pointer to finest level
	    delete scalPhysBCPtr;

	    thisNSPtr = this;
	    while (!thisNSPtr->finestLevel()) {
	      thisNSPtr = thisNSPtr->finerNSPtr();
	    }
	    assert(thisNSPtr->m_level == finest_level);
	    for (int lev=finest_level; lev>=m_level; lev--) {
	      LevelData<FArrayBox>& levelScal = thisNSPtr->newScal(scalComp);
	      LevelData<FArrayBox>& levelCorr = *(scalRefluxRHS[lev]);
	      DataIterator levelDit = levelCorr.dataIterator();
	      for (levelDit.reset(); levelDit.ok(); ++levelDit) {
		levelScal[levelDit()] += levelCorr[levelDit()];
	      }
	      
	      // if a finer level exists, do avgDown as well
	      if (lev < finest_level) {
		AMRNavierStokes& fineNS = *(thisNSPtr->finerNSPtr());
		// quick sanity check
		assert (fineNS.m_level == lev+1);
		CoarseAverage& scalAvgDown = fineNS.m_coarse_average_lambda;
		LevelData<FArrayBox>& fineScal = fineNS.newScal(scalComp);
		scalAvgDown.averageToCoarse(levelScal, fineScal);
	      }
	      thisNSPtr = thisNSPtr->crseNSPtr();
	    } // end loop over levels
	    
	  }

#ifdef BOUSSINESQ
	  // this is as good a place as any to dump out the temperature...
	  if ((scalComp == THETA_COMP) && s_boussinesq && (m_level == 0)) {
	    for (int point=0; point<m_dataPoints.size(); point++) {
	      Real pointTheta = s_bogus_value;
	      
	      thisNSPtr = this;
	      for (int lev=0; lev<=finest_level; lev++) {
		Real levelDx = thisNSPtr->m_dx;
		IntVect levelPoint;
		for (int dir=0; dir<SpaceDim; dir++) {
		  levelPoint[dir] = int(m_dataPoints[point][dir]/levelDx);
		}

		// now loop over grids to find point
		LevelData<FArrayBox> & levelTheta 
		  = thisNSPtr->newScal(THETA_COMP);
		DataIterator levelDit = levelTheta.dataIterator();
		DisjointBoxLayout levelGrids = levelTheta.getBoxes();
		for (levelDit.reset(); levelDit.ok(); ++levelDit) {
		  if (levelGrids.get(levelDit()).contains(levelPoint)) {
		    pointTheta = levelTheta[levelDit()](levelPoint);
		    
		    // check to see if this point is covered by a 
		    // finer grid
		    if (!thisNSPtr->finestLevel()) {
		      const DisjointBoxLayout& finerGrids 
			= thisNSPtr->finerNSPtr()->newScal(THETA_COMP).getBoxes();
		      LayoutIterator fineLit = finerGrids.layoutIterator();
		      int nRefFine = thisNSPtr->refRatio();
		      for (fineLit.reset(); fineLit.ok(); ++fineLit) {
			Box finerBox(finerGrids[fineLit()]);
			finerBox.coarsen(nRefFine);
			// if this level's point is overlain
			// by finer level, reset value to bogus
			// value
			if (finerBox.contains(levelPoint)) {
			  pointTheta = s_bogus_value;
			} 
		      } // end loop over finer-level boxes
		    } // end if a finer level exists			
		  } // end if point is in this grid
		} // end loop over grids on this level

		thisNSPtr = thisNSPtr->finerNSPtr();
	      } // end loop over levels

	      if (pointTheta != s_bogus_value) {
		pout() << "Time: " << m_time << " Theta at point " 
		     << point << " = " << pointTheta << endl;
	      }
	    } // end loop over points
	    
	    
	  } // end if looking for point data
#endif // BOUSSINESQ

	} // end loop over scalar components

	// clean up temporary scalar storage
	for (int lev=0; lev<=finest_level; lev++) {
	  if (scalRefluxRHS[lev] != NULL) {
	    delete scalRefluxRHS[lev];
	    scalRefluxRHS[lev] = NULL;
	  }

	  if (scalRefluxCorr[lev] != NULL) {
	    delete scalRefluxCorr[lev];
	    scalRefluxCorr[lev] = NULL;
	  }
	}
      } // end implicit scalar refluxing

      // now allocate container for composite velocity and lambda
      Vector< LevelData<FArrayBox>* > compVel(finest_level+1,NULL);
      Vector< LevelData<FArrayBox>* > compLambda(finest_level+1,NULL);


      if (m_level > 0) { 
	// coarser level will be coarse-fine BC
	AMRNavierStokes* crseAMRNavierStokesPtr = crseNSPtr();
	DisjointBoxLayout crseGrids = crseAMRNavierStokesPtr->newVel().getBoxes(); 

	LevelData<FArrayBox>* crseVelPtr =
	  new LevelData<FArrayBox>(crseGrids, SpaceDim);	

	crseAMRNavierStokesPtr->velocity(*crseVelPtr, m_time);		
	compVel[m_level-1] = crseVelPtr;

	LevelData<FArrayBox>* crseLambdaPtr = 
	  new LevelData<FArrayBox>(crseGrids, 1);

	crseAMRNavierStokesPtr->lambda(*crseLambdaPtr, m_time);
	compLambda[m_level-1] = crseLambdaPtr;

      }	

      thisNSPtr = this;
      // now loop over levels and construct composite vel, lambda
      for (int lev=m_level; lev <= finest_level; lev++) {
	compVel[lev] = thisNSPtr->m_vel_new_ptr;
	compLambda[lev] = thisNSPtr->m_lambda_new_ptr;
	velBCPtr->applyBCs(*compVel[lev], thisNSPtr->problemDomain(),
			   thisNSPtr->m_dx);
	thisNSPtr = thisNSPtr->finerNSPtr();
      }

      // before we do sync projection, do implicit refluxing,
      // if appropriate
      if (s_reflux_momentum && s_implicit_reflux) {
	// loop over levels and compute RHS
	Vector<LevelData<FArrayBox>* > refluxRHS(finest_level+1,NULL);
	Vector<LevelData<FArrayBox>* > refluxCorr(finest_level+1,NULL);
	// this is necessary because while solver only can do
	// one component, levelfluxRegister::reflux can only
	// do all of them at once.  
	Vector<LevelData<FArrayBox>* > tempRefluxData(finest_level+1,NULL);
	Vector<DisjointBoxLayout> AMRgrids(finest_level+1);
	Vector<int> amrRefRatios(finest_level+1);
	Vector<Real> AmrDx(finest_level+1);
	Vector<ProblemDomain> AmrDomains(finest_level+1);
       
	// loop over levels, allocate storage, set up for AMRSolve
	thisNSPtr = this;
	int startLev=m_level;
	// if crser level exists, define it as well for BC's
	if (startLev > 0) {
	  startLev = startLev-1;
	  thisNSPtr = thisNSPtr->crseNSPtr();
	}
	for (int lev=startLev; lev<=finest_level; lev++) {
	  const DisjointBoxLayout& levelGrids = thisNSPtr->newVel().getBoxes();
	  // recall that AMRSolver can only do one component at a time.
	  // rhs has no ghost cells
	  refluxRHS[lev] = new LevelData<FArrayBox>(levelGrids, 1);
	  tempRefluxData[lev] = new LevelData<FArrayBox>(levelGrids,SpaceDim);
	  //soln has one layer of ghost cells
	  IntVect ghostVect(D_DECL(1,1,1));
	  refluxCorr[lev] = new LevelData<FArrayBox>(levelGrids,1,ghostVect);
	  AMRgrids[lev] = levelGrids;
	  amrRefRatios[lev] = thisNSPtr->refRatio();
	  AmrDx[lev] = thisNSPtr->Dx();
	  AmrDomains[lev] = thisNSPtr->problemDomain();

	  // initialize rhs to 0
	  DataIterator levelDit = tempRefluxData[lev]->dataIterator();
	  LevelData<FArrayBox>& levelRefluxData = *(tempRefluxData[lev]);
	  for (levelDit.reset(); levelDit.ok(); ++levelDit) {
	    levelRefluxData[levelDit()].setVal(0.0);
	  }

	  // while we're here, do refluxing. 
          // (recall that startLev may be coarser than m_level for BC's,
          // however, don't want to do anything to that level)
	  if  ((lev >= m_level) && (lev < finest_level)) {
	    refluxScale = 1.0/AmrDx[lev];
	    thisNSPtr->m_flux_register.reflux(levelRefluxData, refluxScale);
	  }
	    
	  thisNSPtr = thisNSPtr->finerNSPtr();
	}
	
	// coarse BC is 0
	if (m_level > 0) {
	  LevelData<FArrayBox>& crseBCData = *refluxCorr[m_level-1];
	  DataIterator crseDit = crseBCData.dataIterator();
	  for (crseDit.reset(); crseDit.ok(); ++crseDit) {
	    crseBCData[crseDit()].setVal(0.0);
	  }
	}

	// now loop over directions
	Interval solverComps(0,0);
	for (int dir=0; dir<SpaceDim; dir++) {
	  Interval velComps(dir,dir);
	  for (int lev=m_level; lev<=finest_level; ++lev) {	    
	    // copy rhs to single-component holder
	    tempRefluxData[lev]->copyTo(velComps,*(refluxRHS[lev]),
					solverComps);
	    // initial guess for correction is RHS.
	    LevelData<FArrayBox>& levelCorr = *(refluxCorr[lev]);
	    DataIterator levelDit = levelCorr.dataIterator();
	    for (levelDit.reset(); levelDit.ok(); ++levelDit) {	      
	      levelCorr[levelDit()].setVal(0.0);
	    }
	    refluxRHS[lev]->copyTo(solverComps,*(refluxCorr[lev]),
				   solverComps);
	  } // end set initial guess

	  // now set up solver
	  HelmholtzOp viscousOp;
	  const DomainGhostBC* thisBCPtr = m_physBCPtr->viscousRefluxBC(dir);
	  viscousOp.setDomainGhostBC(*thisBCPtr);
	  Real beta = -s_nu*m_dt;
	  viscousOp.setHelmCoeff(beta);
	  delete thisBCPtr;
	  
	  // now define multilevel solver for this component's viscous solve
	  AMRSolver viscousSolver(AMRgrids, AmrDomains,
				  AmrDx, amrRefRatios, finest_level+1,
				  m_level, &viscousOp);
	  
	  if (s_verbosity >= 3) {
	    viscousSolver.setVerbose(true);
	  } else {
	    viscousSolver.setVerbose(false);
	  }
	  viscousSolver.setTolerance(s_viscous_solver_tol);

	  // now solve
	  viscousSolver.solveAMR(refluxCorr, refluxRHS);

	  // now increment velocity with reflux correction
	  for (int lev=m_level; lev<=finest_level; ++lev) {
	    LevelData<FArrayBox>& levelVel = *(compVel[lev]);
	    LevelData<FArrayBox>& levelCorr = *(refluxCorr[lev]);
	    DataIterator levelDit = levelCorr.dataIterator();
	    for (levelDit.reset(); levelDit.ok(); ++levelDit) {
	      levelVel[levelDit()].plus(levelCorr[levelDit()],0,dir,1);
	    }
	  }
	
	} // end loop over directions
	

	// clean up storage
	for (int lev=startLev; lev<=finest_level; lev++) {
	  if (refluxRHS[lev] != NULL) {
	    delete refluxRHS[lev];
	    refluxRHS[lev] = NULL;
	  }

	  if (refluxCorr[lev] != NULL) {
	    delete refluxCorr[lev];
	    refluxCorr[lev] = NULL;
	  }

	  if (tempRefluxData[lev] != NULL) {
	    delete tempRefluxData[lev];
	    tempRefluxData[lev] = NULL;
	  }
	  
	}

      } // end implicit reflux      

      // re-apply physical boundary conditions
      thisNSPtr = this;
      for (int lev = m_level; lev < finest_level; lev++) {	
	velBCPtr->applyBCs(thisNSPtr->newVel(), 
		       thisNSPtr->problemDomain(), 
		       thisNSPtr->m_dx);      
	thisNSPtr = thisNSPtr->finerNSPtr();
      }


      // call projection-based sync operations
      m_projection.doSyncOperations(compVel, compLambda, m_time, m_dt);


      // re-apply physical boundary conditions
      thisNSPtr = this;
      for (int lev = m_level; lev < finest_level; lev++) {	
	velBCPtr->applyBCs(thisNSPtr->newVel(), 
		       thisNSPtr->problemDomain(), 
		       thisNSPtr->m_dx);      
	thisNSPtr = thisNSPtr->finerNSPtr();
      }

      // this is a diagnostic -- compute max(divergence)
      // also compute max(theta) and max(vorticity)
      // and integrals of kinetic energy and vorticity.
      if (m_level == 0) {
	LevelData<FArrayBox>* crseVelPtr = NULL;
	LevelData<FArrayBox>* fineVelPtr = NULL;
	int nRefCrse = -1;
	int nRefFine = -1;

	AMRNavierStokes* thisLevelPtr = this;

	Real maxDiv = 0.0;
        Real maxVort = 0.0;
        Real maxLambdaErr = 0.0;
        Real integralKE = 0.0;

#ifdef BOUSSINESQ
	Real maxTheta = 0.0;
#endif

	for (int lev=m_level; lev <= finest_level; lev++) {
	  LevelData<FArrayBox>& levelVel = *compVel[lev];
	  const DisjointBoxLayout& levelGrids = levelVel.getBoxes();
	  LevelData<FArrayBox> levelDiv(levelGrids, 1);
          LevelData<FArrayBox> levelKE(levelGrids,1);

          LevelData<FArrayBox> levelLambda(levelGrids,1);
          thisLevelPtr->newLambda().copyTo(levelLambda.interval(),
                                           levelLambda, 
                                           levelLambda.interval());

          // since we want max(error(lambda)), subtract 1...
          DataIterator levelDit = levelLambda.dataIterator();
          for (levelDit.begin(); levelDit.ok(); ++levelDit) 
            {
              levelLambda[levelDit()] -= 1.0;
            }

          int numVortComps = 1;
          if (SpaceDim == 3) numVortComps = SpaceDim;
          LevelData<FArrayBox> levelVort(levelGrids, numVortComps);

	  const ProblemDomain& levelDomain = thisLevelPtr->problemDomain();

	  if (lev < finest_level) {
	    fineVelPtr = compVel[lev+1];
	    nRefFine = m_ref_ratio;
	  } else {
	    fineVelPtr = NULL;
	    nRefFine = -1;
	  }

	  compDivergenceCC(levelDiv, levelVel, crseVelPtr, fineVelPtr,
			   thisLevelPtr->m_dx, nRefCrse, nRefFine, 
			   levelDomain, true);
          
          thisLevelPtr->computeVorticity(levelVort);
          thisLevelPtr->computeKineticEnergy(levelKE);

	  DataIterator dit = levelDiv.dataIterator();
	  Real levelMax = 0.0;
          Real levelMaxVort = 0.0;
          Real levelMaxLambda = 0.0;
          Real levelSumKE = 0.0;
	  Interval divComps(0,0);
          Interval vortComps(0,numVortComps-1);

#ifdef BOUSSINESQ
	  Real levelMaxTheta = 0.0;
	  Interval thetaComps(0,0);
#endif

	  if (fineVelPtr != NULL) {
	    const DisjointBoxLayout* finerGridsPtr = &fineVelPtr->getBoxes();
	    levelMax = computeNorm(levelDiv, finerGridsPtr, 
				   nRefFine, thisLevelPtr->m_dx,
				   divComps, 0);

            levelMaxVort = computeNorm(levelVort, finerGridsPtr,
                                       nRefFine, thisLevelPtr->m_dx,
                                       vortComps,0);

            levelMaxLambda = computeNorm(levelLambda, finerGridsPtr,
                                         nRefFine, thisLevelPtr->m_dx,
                                         levelLambda.interval(), 0);

            levelSumKE = computeSum(levelKE, finerGridsPtr, nRefFine,
                                    thisLevelPtr->m_dx, levelKE.interval());
                                   

#ifdef BOUSSINESQ
	    if (s_num_scal_comps > 0) {
	      LevelData<FArrayBox>& levelTheta 
		= thisLevelPtr->newScal(THETA_COMP);
	      levelMaxTheta = computeNorm(levelTheta, finerGridsPtr,
					  nRefFine, thisLevelPtr->m_dx,
					  thetaComps,0);
	    }
#endif
	  } else {
	    const DisjointBoxLayout* finerGridsPtr = NULL;
	    levelMax = computeNorm(levelDiv, finerGridsPtr, 
				   nRefFine, thisLevelPtr->m_dx,
				   divComps, 0);

            levelMaxVort = computeNorm(levelVort, finerGridsPtr,
                                       nRefFine, thisLevelPtr->m_dx,
                                       vortComps, 0);

            levelMaxLambda = computeNorm(levelLambda, finerGridsPtr,
                                         nRefFine, thisLevelPtr->m_dx,
                                         levelLambda.interval(), 0);
                                    
            levelSumKE = computeSum(levelKE, finerGridsPtr, nRefFine,
                                    thisLevelPtr->m_dx, levelKE.interval());

	    
#ifdef BOUSSINESQ
	    if (s_num_scal_comps) {
	      LevelData<FArrayBox>& levelTheta 
		= thisLevelPtr->newScal(THETA_COMP);
	      
	      levelMaxTheta = computeNorm(levelTheta, finerGridsPtr, 
					  nRefFine, thisLevelPtr->m_dx,
					  thetaComps, 0);
	    }
#endif
	  }
	  if (levelMax > maxDiv) maxDiv = levelMax;
          if (levelMaxVort > maxVort) maxVort = levelMaxVort;
          if (levelMaxLambda > maxLambdaErr) maxLambdaErr = levelMaxLambda;
          integralKE += levelSumKE;

#ifdef BOUSSINESQ
	  if (levelMaxTheta > maxTheta) maxTheta = levelMaxTheta;
#endif
	  
	  // now index to next level	  
	  nRefCrse = nRefFine;
	  crseVelPtr = &levelVel;
	  thisLevelPtr = thisLevelPtr->finerNSPtr();	  
	}
	
        // To make it easier to spot differences between runs, 
        // try to print more siginificant digits of certain
        // values to screen.
	//pout() << "Time = " << m_time << " Max Div(u) = " << maxDiv << endl;
	//pout() << "Time = " << m_time << " Max(Vorticity) = " << maxVort << endl;
        //pout() << "Time = " << m_time << " Max(Err(Lambda) = " << maxLambdaErr << endl;
        //pout() << "Time = " << m_time << " Integral(Kinetic Energy) = " 
        //     << integralKE << endl;

        pout() << setiosflags(ios::scientific) << setprecision(15);
	pout() << "Time = " << setw(15) << m_time 
               << setw(30) << " Max Div(u) = " 
               << setw(23)  << maxDiv << endl;
	pout() << "Time = " << setw(15) << m_time 
               << setw(30) << " Max(Vorticity) = " 
               << setw(23) << maxVort  << endl;
        pout() << "Time = " << setw(15) << m_time 
               << setw(30) << " Max(Err(Lambda) = " 
               << setw(23) << maxLambdaErr  << endl;        
        pout() << "Time = " << setw(15) << m_time 
               << setw(30) << " Integral(Kinetic Energy) = " 
               << setw(23) << integralKE << endl;

#ifdef BOUSSINESQ
	if (s_num_scal_comps > 0) {
	  pout() << "Max Theta = " << maxTheta << endl;
	}
#endif
      }	  
	
      if (m_level > 0) {
	// need to clean up storage
	if (compLambda[m_level-1] != NULL) {
	  delete compLambda[m_level-1];
	  compLambda[m_level-1] = NULL;
	}

	if (compVel[m_level-1] != NULL) {
	  delete compVel[m_level-1];
	  compVel[m_level-1] = NULL;
	}
      }
    

    } // end sync projection operations

    delete velBCPtr;

    // end if not finest level
  } else {

    // in case this is the only level
    if (m_level == 0) {

      // create boundary condition object and set physical BC's
      bool isViscous = (s_nu > 0.0);
      basicVelBC* velBCPtr = m_physBCPtr->uStarBC(isViscous);
      velBCPtr->applyBCs(newVel(), m_problem_domain, m_dx);

      LevelData<FArrayBox>* crseVelPtr = NULL;
      LevelData<FArrayBox>* fineVelPtr = NULL;
      int nRefCrse = -1;
      int nRefFine = -1;
      
      Real maxDiv = 0.0;

      const DisjointBoxLayout& grids = newVel().getBoxes();
      LevelData<FArrayBox> thisDiv(grids, 1);
#ifdef BOUSSINESQ
      Real maxTheta = 0.0;
      LevelData<FArrayBox>& theta = newScal(THETA_COMP);
#endif

      Interval velComps = newVel().interval();
      newVel().exchange(velComps);

      compDivergenceCC(thisDiv, newVel(), crseVelPtr, fineVelPtr,
		       m_dx, nRefCrse, nRefFine, m_problem_domain,
		       true);

      DataIterator dit=thisDiv.dataIterator();
      Interval divComps(0,0);

      const DisjointBoxLayout* finerGridsPtr = NULL;
      maxDiv = computeNorm(thisDiv, finerGridsPtr, nRefFine,
			   m_dx, divComps, 0);
#ifdef BOUSSINESQ
      Interval thetaComps(0,0);
      maxTheta = computeNorm(theta, finerGridsPtr, nRefFine,
			     m_dx, thetaComps, 0);
#endif

      pout() << "Time = " << m_time << " Max Div(u) = " << maxDiv << endl;
#ifdef BOUSSINESQ
      pout() << "Max Theta = " << maxTheta << endl;
#endif

      delete velBCPtr;
    } // end if single-grid case...
    
  } // end if finest level      

}
