/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include "AMRNavierStokesHeaders.H"

void 
AMRNavierStokes::predictVelocities(LevelData<FArrayBox>& a_uDelU,
			    LevelData<FluxBox>& a_advVel)
{
  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::predictVelocities: " << m_level << endl;
  }

  bool isViscous = (s_nu > 0.0);

  const DisjointBoxLayout& levelGrids = a_advVel.getBoxes();
  
  // will need edge-centered storage for all velocity components
  LevelData<FluxBox> uHalf(levelGrids, SpaceDim);

  // will also need grad(eLambda) and grad(phi_mac)
  LevelData<FluxBox>& grad_eLambda = m_projection.grad_eLambda();
  LevelData<FluxBox> gradPhi(levelGrids, SpaceDim);
  m_projection.gradPhi(gradPhi);


  // for tracing, will need to fill in boundary values for 
  // grown copy of velocity
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  LevelData<FArrayBox> traceVel(levelGrids, SpaceDim, ghostVect);
  LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim, ghostVect);
  LevelData<FArrayBox>* crseVelPtr = NULL;

  if (s_set_bogus_values) {
    DataIterator dit = traceVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      uHalf[dit()].setVal(s_bogus_value);
      traceVel[dit()].setVal(s_bogus_value);
      viscousSource[dit()].setVal(s_bogus_value);
    }
  }

  // now fill in velocity field
  // m_time is new time
  Real old_time = m_time - m_dt;
  fillVelocity(traceVel, old_time);

  // do physical boundary conditions on traceVel
  basicVelBC* velBCPtr = NULL;

  if (isViscous) {    
    velBCPtr = m_physBCPtr->viscousVelBC();
    velBCPtr->applyBCs(traceVel,m_problem_domain,m_dx);
    delete velBCPtr;

    // if crse level exists, fill coarse velocity BC for laplacian
    if (m_level > 0) {
      const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
      crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
      crseNSPtr()->fillVelocity(*crseVelPtr, old_time);
    }
    
    computeLapVel(viscousSource, traceVel, crseVelPtr);

    DataIterator dit = viscousSource.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].mult(s_nu);
    }
    
    // this is a clunky way to reset C/F BC's to
    // conservative linear interpolation...
    fillVelocity(traceVel, old_time);

  } else {
    // inviscid case
    DataIterator dit = viscousSource.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].setVal(0.0);
    }
  }

  // tracing uses inviscid BC's
  velBCPtr = m_physBCPtr->tracingVelBC();
  velBCPtr->applyBCs(traceVel, m_problem_domain, m_dx);
  delete velBCPtr;
  
  
#ifdef BOUSSINESQ  
  // if appropriate, add in forcing due to theta
  if (s_boussinesq) {
    LevelData<FArrayBox>& oldTheta = oldScal(THETA_COMP);
    // "up" direction is always the last one
    int thetaDir = SpaceDim -1;
    // loop over grids
    DataIterator dit = viscousSource.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      assert(s_theta_coeff == 1.0);
      viscousSource[dit()].plus(oldTheta[dit()],0,
				thetaDir,1);
    }
  }
#endif
    



  // loop over velocity components
  for (int velComp = 0; velComp < SpaceDim; velComp++) {
    // need to compute predicted edge-centered values for each
    // velocity component in each coordinate direction. begin
    // loop over coordinate directions 
    for (int dir=0; dir<SpaceDim; dir++) {
      // if this is a normal edge, we already have edge-centered
      // predicted values, which are divergence-free (the advection
      // velocities).  treat this case differently
      if (dir == velComp) {
	DataIterator dit = uHalf.dataIterator();
	for (dit.reset(); dit.ok(); ++dit) {

	  // copy advVel -> uHalf
	  // srcComp is always 0, since advVel only has one component
	  int  srcComp = 0;
	  int  destComp = velComp;
	  //a_advVel.copyTo(srcComp, uHalf, destComp);
	  FluxBox& thisVel = uHalf[dit()];
	  FArrayBox& thisVelDir = thisVel[dir];

	  FluxBox& thisAdvVel = a_advVel[dit()];
	  FArrayBox& thisAdvVelDir = thisAdvVel[dir];

	  thisVelDir.copy(thisAdvVelDir,srcComp, destComp, 1);
	  
          if (s_applyFreestreamCorrection) 
            {
              // now need to subtract off grad(eLambda)
              // this is the due to the difference between _advecting_
              // and _advected_ velocities
              
              FluxBox& thisCorr = grad_eLambda[dit()];
              
              FArrayBox& thisCorrDir = thisCorr[dir];
              
              thisVelDir.minus(thisCorrDir, 0, dir, 1);
            }
	}
      } else {
	// if not a normal direction, need to perform tracing
	Interval srcComps(velComp, velComp);
	Interval destComps(velComp, velComp);

	DataIterator dit = uHalf.dataIterator();
	for (dit.reset(); dit.ok(); ++dit) {
	  // const Box& gridBox = levelGrids[dit()];
	  //const Box& grownBox = traceVel[dit()].box();
	  FArrayBox& srcFab = viscousSource[dit()];

	  FluxBox& thisUhalf = uHalf[dit()];
	  FArrayBox& thisUHalfDir = thisUhalf[dir];

	  FluxBox& thisAdvVel = a_advVel[dit()];
	  FArrayBox& thisOldVel = traceVel[dit()];

	  TraceState(thisUHalfDir, thisOldVel, thisOldVel, 
		     thisAdvVel, srcFab, problemDomain(),
		     levelGrids[dit()], m_dt, m_dx, dir,
		     srcComps, destComps);


	// need to add MAC correction to traced velocities
	thisUHalfDir.minus(gradPhi[dit()][dir], velComp, 
			   velComp, 1);

	}


      }
    } // end loop over tracing directions
  } // end loop over velocity components
  
  // now loop over grids to compute advection and update flux registers
  DataIterator dit = a_uDelU.dataIterator();
  
  // for nonconservative update, also need a cell-centered
  // advection velocity -- get this by averaging edge-centered
  // advection velocity to cell centers.  this one shouldn't need
  // any ghost cells.

  LevelData<FArrayBox> cellAdvVel(levelGrids, SpaceDim);
  EdgeToCell(a_advVel, cellAdvVel);

  for (dit.reset(); dit.ok(); ++dit) {
    FluxBox& thisUHalf = uHalf[dit()];
    FArrayBox& thisCellAdvVel = cellAdvVel[dit()];
    const Box& thisBox = levelGrids[dit()];
    FArrayBox& this_uDelU = a_uDelU[dit()];
    this_uDelU.setVal(0.0);
    
    // to do this in a dimensionality independent way, 
    // loop over directions.
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& uHalfDir = thisUHalf[dir];
     
      // this will increment this_uDelU with
      // cellVel*d(uHalfDir)/d[xyz] for all velocity components
      FORT_UDELS(CHF_FRA(this_uDelU),
		 CHF_FRA(uHalfDir),
		 CHF_FRA(thisCellAdvVel),
		 CHF_BOX(thisBox),
		 CHF_REAL(m_dx),
		 CHF_INT(dir));
    }

    // now need to do flux register stuff
    // first compute fluxes (u_edge*u_adv)
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& this_uEdge = uHalf[dit()][dir];
      FArrayBox& thisAdvVel = a_advVel[dit()][dir];

      // now need to loop over traced velocity components
      for (int velComp=0; velComp<SpaceDim; ++velComp) {
	this_uEdge.mult(thisAdvVel,0,velComp,1);
      }
    }

    Interval velComps(0,SpaceDim-1);
    if (!finestLevel()) {
      assert (m_flux_register.isDefined());
      for (int dir=0; dir < SpaceDim; dir++) {
	// because we eventually want -D_R(U-del-U), signs
	// on fluxMult are opposite of what you might expect...
	Real fluxMult = -m_dt;
	// debugging change
	//fluxMult = 0.0;
	m_flux_register.incrementCoarse(uHalf[dit()][dir],
					fluxMult, dit(), velComps,
					velComps, dir);
      }
    }
    
    if (m_level > 0) {
      LevelFluxRegister& crseFR = crseNSPtr()->m_flux_register;
      assert(crseFR.isDefined());
      // because we will eventually want -D_R(U-del-U) signs
      // on fluxMult are opposite of what you might expect...      
      Real fluxMult = -m_dt;
      // debugging change
      //fluxMult = 0.0;
      for (int dir=0; dir<SpaceDim; dir++) {
	crseFR.incrementFine(uHalf[dit()][dir],
			     fluxMult, dit(), velComps,
			     velComps, dir, Side::Lo);

	crseFR.incrementFine(uHalf[dit()][dir],
			     fluxMult, dit(), velComps,
			     velComps, dir, Side::Hi);
      
      }
    }


  } // end loop over grids 


  // at this point, we've computed the advective term, and updated
  // the flux registers with the momentum flux 


  if (crseVelPtr != NULL) { 
    delete crseVelPtr;
    crseVelPtr = NULL;
  }

}

// the approximation to u-dot-del-u is passed into this
// function in a_uStar -- at end of function, a_uStar
// will contain approximation to new velocity:
// uStar + dt*gradPi
void
AMRNavierStokes::computeUStar(LevelData<FArrayBox>& a_uStar)
{

  bool isViscous = (s_nu > 0.0);
  Interval velComps(0,SpaceDim-1);

  if (!isViscous) {
    /// compute uStar = oldVel - dt*(u-dot-del-u)
    DataIterator dit = a_uStar.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      Real mult = -1.0*m_dt;
      a_uStar[dit()] *= mult;
      a_uStar[dit()] += oldVel()[dit()];
    }
  } else { 
    // viscous case is more elaborate -- need to do viscous solves.


    const DisjointBoxLayout& levelGrids = a_uStar.getBoxes();

    // these are parameters for TGA viscous solves
    Real a = 2.0 - sqrt(2.0) - TGA_EPS;
    Real discr = sqrt(a*a - 4.0*a + 2.0);
    Real r1 = (2.0*a - 1.0)/(a + discr);
    Real r2 = (2.0*a - 1.0)/(a - discr);  
    
    // m_time is new time
    Real old_time = m_time - m_dt;

    // define temporary storage
    LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim);
    LevelData<FArrayBox> tempStorage(levelGrids, SpaceDim, 
				     IntVect::TheUnitVector());
    LevelData<FArrayBox> gradPi(levelGrids, SpaceDim);
    LevelData<FArrayBox>* crseDataPtr = NULL;    

    LevelData<FluxBox> viscousFlux(levelGrids, SpaceDim);
    DataIterator dit = a_uStar.dataIterator();

    // these are necessary because solver can only handle one component
    // at a time.
    LevelData<FArrayBox> solverRHS(levelGrids, 1, 
				   IntVect::TheUnitVector());
    LevelData<FArrayBox> intermediateVel(levelGrids,1,
					 IntVect::TheUnitVector());
    LevelData<FluxBox> tempFlux(levelGrids, 1);

    LevelData<FArrayBox>* crseSolverDataPtr = NULL;

    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
	viscousSource[dit()].setVal(s_bogus_value);
	tempStorage[dit()].setVal(s_bogus_value);
	solverRHS[dit()].setVal(s_bogus_value);
	intermediateVel[dit()].setVal(s_bogus_value);
	for (int dir=0; dir<SpaceDim; ++dir) {
	  viscousFlux[dit()][dir].setVal(s_bogus_value);
	  tempFlux[dit()][dir].setVal(s_bogus_value);
	}
      }
    }	

    // will need single-component QuadCFInterp to do fluxes,
    // and Laplacians
    QuadCFInterp singleCompCFInterp;

    if (m_level > 0) {
      const DisjointBoxLayout& crseBoxes = crseNSPtr()->newVel().getBoxes();
      crseDataPtr = new LevelData<FArrayBox>(crseBoxes, SpaceDim);
      crseSolverDataPtr = new LevelData<FArrayBox>(crseBoxes, 1);
      int nRefCrse = crseNSPtr()->refRatio();
      singleCompCFInterp.define(levelGrids, &crseBoxes, m_dx, 
				nRefCrse, 1, m_problem_domain);
    }

    //Real viscousBeta; //not used?  (ndk)
    if (s_viscous_solver_type == TGA) {
      //viscousBeta = -r2*s_nu*m_dt; //not used? (ndk)
    } else {
      MayDay::Error("undefined viscous solver type");
    }

    // first get gradPi
    m_projection.gradPi(gradPi);
    
    // "f" is -u-del-u -gradPi +source
    for (dit.reset(); dit.ok(); ++dit) {
      a_uStar[dit()] += gradPi[dit()];
      a_uStar[dit()] *= -1.0;
    }

    // if necessary, add in source term here
    // use averaged old- and new-time theta
    // to approximate half-time theta
#ifdef BOUSSINESQ
    if (s_boussinesq) {
      LevelData<FArrayBox>& oldTheta = oldScal(THETA_COMP);
      LevelData<FArrayBox>& newTheta = newScal(THETA_COMP);
      Interval thetaComps(0,0);
      oldTheta.copyTo(thetaComps, tempStorage, thetaComps);
      for (dit.reset(); dit.ok(); ++dit) {
	int thetaDir = SpaceDim -1;
	tempStorage[dit()].plus(newTheta[dit()],0,0,1);
	tempStorage[dit()].mult(0.5*s_theta_coeff,0,1);
	a_uStar[dit()].plus(tempStorage[dit()],0,thetaDir,1);
      }
      
    }
#endif
      


    // now compute diffused version of this
    if (crseDataPtr != NULL) {
      DataIterator crseDit = crseDataPtr->dataIterator();
      for (crseDit.reset(); crseDit.ok(); ++crseDit) {
	((*crseDataPtr)[crseDit()]).setVal(0.0);
      }
    }

#ifdef MATCH_OLDCODE
    // this is a kluge to make new code agree with old code
    // C/F BC's are different -- here we can force them
    // to be extrap by being clever
    if (m_level > 0) {
      // this BC object will contain only extrapBC's 
      // and will be used at C/F interfaces
      DomainGhostBC bogusBC;
      // loop over directions
      for (int dir=0; dir<SpaceDim; dir++) {
	// loop over hi-lo
	SideIterator sit;
	for (sit.begin(); sit.ok(); ++sit) {
	  HOExtrapBC thisBC(dir,sit(), velComps);
	  bogusBC.setBoxGhostBC(thisBC);
	}
      }
    
      // now loop over grids
      for (dit.reset(); dit.ok(); ++dit) {
	Box thisGridBox = levelGrids[dit()];
	FArrayBox& thisData = a_uStar[dit()];
	bogusBC.applyHomogeneousBCs(thisData,thisGridBox,m_dx);
      }
      
    } // at this point, ALL ghost cells are set by extrap      	
    
    // now set physical BC's on (-uDelU-GradPi)
    bool isViscous = (s_nu > 0.0);
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC(isViscous);

#ifdef MATCH_OLDCODE
    //velBC.isViscous(false);
#endif

    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);    
    a_uStar.exchange(velComps);
    delete velBCPtr;

    // this computes L(-uDelU-GradPi) and puts it in tempStorage

    for (dit.reset(); dit.ok(); ++dit) {
      FORT_OPERATORLAP(CHF_FRA(tempStorage[dit()]),
		       CHF_FRA(a_uStar[dit()]),
		       CHF_BOX(levelGrids.get(dit())),
		       CHF_CONST_REAL(m_dx));
    }


    {
      // put these in their own localized scope
      viscousBC viscBC;
      viscBC.applyBCs(tempStorage,m_problem_domain, m_dx);
      tempStorage.exchange(tempStorage.interval());
    }

    // before we add in stuff, compute flux due to antidiffusive part
    // in this case, use the extrap C/F BC we've already set.
    // (this version of the levelGradientMac fn doesn't do C/F BC's)
    levelGradientMAC(viscousFlux, a_uStar, m_dx);


#else    
    cs4.start();

    // set physical BC's on (-uDelU-GradPi)
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC();
    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);    
    a_uStar.exchange(velComps);
    delete velBCPtr;

    // this computes L(-uDelU-GradPi) and puts it in tempStorage
    computeLapVel(tempStorage, a_uStar, crseDataPtr);

    // before we add in stuff, compute flux due to antidiffusive part
    levelGradientMAC(viscousFlux, a_uStar, crseDataPtr, m_dx, 
		     m_velCFInterp);


#endif
    
    // multiply grad(forcing term) by mult, then initialize
    // and increment viscousFlux (will increment flux registers
    // all at once)
    // sign convention will be so that flux will have the opposite
    // sign when incrementing flux registers as the fluxes
    // computed in earlier case.  (same sign convention as in 
    // old code).

    Real mult = (0.5-a)*m_dt*s_nu;
    // this is experimental, for debugging purposes
    //mult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = viscousFlux[dit()];
      for (int dir=0; dir<SpaceDim; dir++) {
	thisFlux[dir] *= mult;
      }
    }
	
    // source term we will use is (I + (half-a)*dt*nu*L)(-uDelU-gradPi)
    // this will be stored in a_uStar
    for (dit.reset(); dit.ok(); ++dit) {
      tempStorage[dit()] *= mult;
      a_uStar[dit()] += tempStorage[dit()];
    }
    
    // now need to get old-time velocity
    fillVelocity(tempStorage, old_time);
    if (crseDataPtr != NULL) {
      crseNSPtr()->fillVelocity(*crseDataPtr, old_time);
    }
    
    velBCPtr = m_physBCPtr->viscousVelBC();
    velBCPtr->applyBCs(tempStorage, m_problem_domain, m_dx);
    delete velBCPtr;

    mult = -(1.0-a)*s_nu;
    // this is experimental, for debugging
    //mult = 0.0;


    // compute viscous flux due to old-time velocity
    {
      // need a second multicomponent LevelData<FluxBox>
      // to hold these fluxes -- goes out of scope and 
      // is deleted when we're done
      LevelData<FluxBox> oldViscFlux(levelGrids, SpaceDim);
      levelGradientMAC(oldViscFlux, tempStorage, crseDataPtr, 
		       m_dx, m_velCFInterp);
            
      for (dit.reset(); dit.ok(); ++dit) {
	FluxBox& thisFlux = oldViscFlux[dit()];
	FluxBox& thisViscFlux = viscousFlux[dit()];
	for (int dir = 0; dir<SpaceDim; dir++) {
	  thisFlux[dir].mult(mult);
	  thisViscFlux[dir] -= thisFlux[dir];
	}
      }
    } // temporary flux container goes out of scope here.
    
    // compute L(old_velocity)
    computeLapVel(viscousSource, tempStorage, crseDataPtr);   


    // now loop over boxes, multiply viscousSource by mult,
    // then add old-time velocity and viscousSource to 
    // the RHS for the first solve, which will all be in
    // a_uStar, and which is 
    // RHS = (I + (1-a)*nu*dt*L)u_n +dt*f,
    // where f = (I +(half-a)*dt*nu*L)(-uDelU - gradPi)

    mult = (1.0-a)*m_dt*s_nu;
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()] *= mult;
      a_uStar[dit()] *= m_dt;
      
      a_uStar[dit()] += viscousSource[dit()];
      a_uStar[dit()] += tempStorage[dit()];
    }

    // set up coarse-level BC
    Real intermediateTime = old_time + (1-r1)*m_dt;
    if (crseDataPtr != NULL) {
      crseNSPtr()->fillVelocity(*crseDataPtr, intermediateTime);
    }


    // finally ready to do intermediate solve!

    a_uStar.exchange(velComps);


    // loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {


      // copy RHS and coarse BC into single-component holder
      LevelData<FArrayBox>& crseBC = *crseSolverDataPtr;
      LevelData<FArrayBox>& crseVel = *crseDataPtr;
      for (dit.reset(); dit.ok(); ++dit) {
	solverRHS[dit()].copy(a_uStar[dit()],dir,0,1);
	// initialize soln to be RHS
	intermediateVel[dit()].copy(a_uStar[dit()],dir,0,1);
      }
      if (crseDataPtr != NULL) {
	DataIterator crseDit = crseVel.dataIterator();
	for (crseDit.reset(); crseDit.ok(); ++crseDit) {
	  // copy coarse vel BC into single-component holder
	  crseBC[crseDit()].copy(crseVel[crseDit()],dir,0,1);
	}
      }
      
      
      mult = -r2*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);            

      DomainGhostBC* thisVelBCPtr = m_physBCPtr->viscousSolveBC(dir);
      m_viscous_solver.setBC(*thisVelBCPtr);
      
      m_viscous_solver.levelSolve(intermediateVel, crseSolverDataPtr,
				  solverRHS, false);    


      // increment fluxes with flux due to intermediate velocity
      // set physical BC's -- this also relies on the fact that
      // we use the same velocity components for all directions
      // (however, need to use the x-direction DomainGhostBC
      // because of the single-component nature of intermediateVel
      
      //const DomainGhostBC& thisVelBC = velBC.DomGhostBC(dir);
      for (dit.reset(); dit.ok(); ++dit) {
	thisVelBCPtr->applyHomogeneousBCs(intermediateVel[dit()], 
					  m_problem_domain, m_dx);
      }
    
      levelGradientMAC(tempFlux, intermediateVel, crseSolverDataPtr, 
		       m_dx, singleCompCFInterp);
      
      mult = -r2*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
	FluxBox& thisFlux = tempFlux[dit()];
	FluxBox& thisViscFlux = viscousFlux[dit()];
	for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
	  thisFlux[fluxDir].mult(mult);
	  thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
	}
      }
      
      // now ready to do second solve
      mult = -r1*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);
      
      
      // fill coarse-level boundary condition...
      if (crseDataPtr != NULL) {
	crseNSPtr()->fillVelocity(*crseSolverDataPtr, m_time,
				  dir, 0, 1);
      }
      
      // initialize initial guess 
      for (dit.reset(); dit.ok(); ++dit) {
	solverRHS[dit()].copy(intermediateVel[dit()]);
      }


      // and solve away
      m_viscous_solver.levelSolve(solverRHS, crseSolverDataPtr, intermediateVel,
				  false);      

    

      // set physical BC's
      for (dit.reset(); dit.ok(); ++dit) {
	thisVelBCPtr->applyHomogeneousBCs(solverRHS[dit()], 
					  m_problem_domain, m_dx);
      }
      
      // compute flux due to uStar
      levelGradientMAC(tempFlux, solverRHS, crseSolverDataPtr, 
		       m_dx, singleCompCFInterp);
      
      mult = -r1*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
	FluxBox& thisFlux = tempFlux[dit()];
	FluxBox& thisViscFlux = viscousFlux[dit()];
	for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
	  thisFlux[fluxDir].mult(mult);
	  thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
	}

	// need to copy computed uStar into a_uStar
	a_uStar[dit()].copy(solverRHS[dit()],0,dir,1);
	
      } // end loop over grids

      delete thisVelBCPtr;

    } // end loop over directions



   
    // now increment flux registers
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisViscFlux = viscousFlux[dit()];
      if (!finestLevel()) {
	assert(m_flux_register.isDefined());
	for (int dir=0; dir<SpaceDim; dir++) {
	  // this is to make things look the same as old code
	  //thisViscFlux[dir] *= m_dx;
	  // use same multipliers as momentum fluxes, but
	  // opposite sign convention
	  Real fluxMult = m_dt;
	  m_flux_register.incrementCoarse(thisViscFlux[dir],
					  fluxMult, dit(), velComps,
					  velComps, dir);
	}
      }
      
      if (m_level > 0) {
	LevelFluxRegister& crseFR = crseNSPtr()->m_flux_register;
	assert(crseFR.isDefined());
	// use same multipliers as momemtum fluxes, but opposite
	// sign convention
	Real fluxMult = m_dt;
	for (int dir=0; dir<SpaceDim; dir++) {
	  // this is to make things look the same as old code
	  //thisViscFlux[dir] *= m_dx;

	  crseFR.incrementFine(thisViscFlux[dir],
			       fluxMult, dit(), velComps, 
			       velComps, dir, Side::Lo);
	  
	  crseFR.incrementFine(thisViscFlux[dir],
			       fluxMult, dit(), velComps, 
			       velComps, dir, Side::Hi);
	}
      }
    
    } // end loop over grids for flux register updates

    // final thing to do -- subtract off dt*gradPi term
    for (dit.reset(); dit.ok(); ++dit) {
      gradPi[dit()] *= m_dt;
      a_uStar[dit()] += gradPi[dit()];
    }
  
    
    // clean up temp storage
    if (crseDataPtr != NULL) {
      delete crseDataPtr;
      crseDataPtr = NULL;
    }
    
    if (crseSolverDataPtr != NULL) {
      delete crseSolverDataPtr;
      crseSolverDataPtr = NULL;
    }

  } // end if viscous

}

