/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// CCProjector.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, January 19, 2000


#include "CCProjector.H"
#include "AMRSolver.H"
#include "Gradient.H"
#include "GradientF_F.H"
#include "Divergence.H"
#include "PoissonOp.H"
#include "CoarseAverageEdge.H"

#include "LayoutIterator.H"
#include "ParmParse.H"
#include "PoissonBC.H"
#include "boxVelBC.H"
#include "HOExtrapBC.H"
#include "scalarDirichletBC.H"

#include "computeSum.H"
#include "computeNorm.H"

#ifdef HDF5
#include "HDF5.H"
#include "AMRIO.H"
#endif

/// define static variables here
bool CCProjector::m_doSyncProjection = true;
bool CCProjector::m_applySyncCorrection = true;
bool CCProjector::s_applyVDCorrection = true;
bool CCProjector::m_doQuadInterp = true;
Real CCProjector::m_etaLambda = 0.9;
bool CCProjector::pp_init = false;
int  CCProjector::s_verbosity = 0;

/// first define quick-n-easy access functions

// ------------------------------------------------------
int 
CCProjector::getLevel() const  {return m_level;}

// ------------------------------------------------------
const ProblemDomain&
CCProjector::dProblem() const {return m_domain;}

// ------------------------------------------------------
const DisjointBoxLayout& 
CCProjector::getBoxes() const {return m_Pi.getBoxes();}

// ------------------------------------------------------
int
CCProjector::nRefCrse() const {return m_nRefCrse;}

// ------------------------------------------------------
Real
CCProjector::dx() const {return m_dx;}

// ------------------------------------------------------
Real
CCProjector::etaLambda() const {return m_etaLambda;}

// ------------------------------------------------------
CCProjector*
CCProjector::fineProjPtr() const  {return m_fineProjPtr;}

// ------------------------------------------------------
CCProjector*
CCProjector::crseProjPtr() const {return m_crseProjPtr;}

// ------------------------------------------------------
LevelData<FArrayBox>& 
CCProjector::phi() {return m_phi;}

// ------------------------------------------------------
const LevelData<FArrayBox>& 
CCProjector::phi() const {return m_phi;}

// ------------------------------------------------------
LevelData<FArrayBox>&
CCProjector::eSync() {return m_eSync;}

// ------------------------------------------------------
const LevelData<FArrayBox>&
CCProjector::eSync() const  {return m_eSync;}

// ------------------------------------------------------
LevelData<FArrayBox>&
CCProjector::Pi() {return m_Pi;}

// ------------------------------------------------------
LevelData<FArrayBox>&
CCProjector::eLambda() {return m_eLambda;}

// ------------------------------------------------------
const LevelData<FArrayBox>&
CCProjector::eLambda() const {return m_eLambda;}

// ------------------------------------------------------
LevelData<FluxBox>&
CCProjector::grad_eLambda() {return m_grad_eLambda;} 

// ------------------------------------------------------
const LevelData<FluxBox>&
CCProjector::grad_eLambda() const {return m_grad_eLambda;} 

// ------------------------------------------------------
bool
CCProjector::doSyncProjection() const {return m_doSyncProjection;}

// ------------------------------------------------------
bool
CCProjector::doMacSync() const {return (m_etaLambda > 0.0);}

// ------------------------------------------------------
bool
CCProjector::isInitialized() const {return m_isInitialized;}

// ------------------------------------------------------
bool
CCProjector::doQuadInterp() const {return m_doQuadInterp;}

// ------------------------------------------------------
QuadCFInterp&
CCProjector::quadCFInterpolator() {return m_cfInterp;}

// ------------------------------------------------------
bool 
CCProjector::isFinestLevel() const {return m_finest_level;}

// ------------------------------------------------------
void
CCProjector::isFinestLevel(bool a_finest_level) 
{
  m_finest_level = a_finest_level;
}

// ------------------------------------------------------
void 
CCProjector::verbosity(int a_verbosity) 
{
  s_verbosity = a_verbosity;
}


// ------------------------------------------------------
int
CCProjector::verbosity() const
{
  return s_verbosity;
}


// ------------------------------------------------------
void
CCProjector::setPhysBC(const PhysBCUtil& a_bc)
{
  if (m_physBCPtr != NULL) {
    delete m_physBCPtr;
    m_physBCPtr = NULL;
  }
  m_physBCPtr = a_bc.newPhysBCUtil();
}

// ------------------------------------------------------
PhysBCUtil*
CCProjector::getPhysBCPtr() const
{
  return m_physBCPtr;
}


// ------------------------------------------------------
// default constructor
CCProjector::CCProjector()
{
  m_level = -1;
  m_nRefCrse = -1;
  m_isInitialized = false;  
  // have to initialize this to _something_!
  m_finest_level = false;
  m_physBCPtr = NULL;
}


// ------------------------------------------------------
// destructor
CCProjector::~CCProjector()
{
  if (m_physBCPtr != NULL) {
    delete m_physBCPtr;
    m_physBCPtr = NULL;
  }

  // everything else should be automatic here
}


// ------------------------------------------------------
CCProjector::CCProjector(const DisjointBoxLayout& a_grids,
			 const Box& a_domain,
			 const Real a_dx,
			 CCProjector* a_finerProjPtr,
			 CCProjector* a_crseProjPtr,
			 int a_nRefCrse, 
			 int a_level,
			 const PhysBCUtil& a_physBC)
{
  m_physBCPtr = NULL;
  ProblemDomain physdomain(a_domain);
  define(a_grids, physdomain, a_dx, a_finerProjPtr, a_crseProjPtr, 
	 a_nRefCrse, a_level, a_physBC);
}


// ------------------------------------------------------
CCProjector::CCProjector(const DisjointBoxLayout& a_grids,
			 const ProblemDomain& a_domain,
			 const Real a_dx,
			 CCProjector* a_finerProjPtr,
			 CCProjector* a_crseProjPtr,
			 int a_nRefCrse, 
			 int a_level,
			 const PhysBCUtil& a_physBC)
{
  m_physBCPtr = NULL;
  define(a_grids, a_domain, a_dx, a_finerProjPtr, a_crseProjPtr, 
	 a_nRefCrse, a_level, a_physBC);
}

// ------------------------------------------------------
void 
CCProjector::define(const DisjointBoxLayout& a_grids,
		    const Box& a_domain,
		    const Real a_dx, 
		    CCProjector* a_finerProjPtr,
		    CCProjector* a_crseProjPtr,
		    int a_nRefCrse,
		    int a_level,
		    const PhysBCUtil& a_physBC)
{
  ProblemDomain physdomain(a_domain);
  define(a_grids, physdomain, a_dx, a_finerProjPtr, a_crseProjPtr, 
	 a_nRefCrse, a_level, a_physBC);  
}



// ------------------------------------------------------
void 
CCProjector::define(const DisjointBoxLayout& a_grids,
		    const ProblemDomain& a_domain,
		    const Real a_dx, 
		    CCProjector* a_finerProjPtr,
		    CCProjector* a_crseProjPtr,
		    int a_nRefCrse,
		    int a_level,
		    const PhysBCUtil& a_physBC)
{

  // set physical boundary condition object
  setPhysBC(a_physBC);
  
  if (!pp_init) {
    variableSetUp();
  }

  assert (a_level > -1);
  m_level = a_level;
  m_dx = a_dx;
  m_domain = a_domain;

  m_fineProjPtr = a_finerProjPtr;

  m_crseProjPtr = a_crseProjPtr;
  if (m_crseProjPtr != NULL) {
    m_nRefCrse = a_nRefCrse;
    m_crseProjPtr->setFineProj(this);
    m_crseProjPtr->isFinestLevel(false);

    const DisjointBoxLayout& crseGrids = a_crseProjPtr->getBoxes();
    m_cfInterp.define(a_grids, &crseGrids, a_dx, a_nRefCrse,
		      1,a_domain);

  }
    

  IntVect ghostVect(IntVect::Unit);

  // define persistent storage
  m_Pi.define(a_grids,1,ghostVect);
  m_phi.define(a_grids,1,ghostVect);
  m_eLambda.define(a_grids,1,ghostVect);
  m_eSync.define(a_grids,1,ghostVect);

  m_grad_eLambda.define(a_grids,1);


  // set grad_e_lambda to be 0
  // also initialize all other fields to a bogus value
  DataIterator dit = m_grad_eLambda.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    m_grad_eLambda[dit()].setVal(0.0);
    m_Pi[dit()].setVal(1.0e5);
    m_phi[dit()].setVal(1.0e5);
    // these need to start as 0 for boundary conditions
    // for first intermediate solve if more than one level
    // of refinement
    m_eLambda[dit()].setVal(0.0);
    m_eSync[dit()].setVal(0.0);
  }

  const DisjointBoxLayout* crseGridsPtr = NULL;
  if (a_crseProjPtr != NULL) {
    crseGridsPtr = &(m_crseProjPtr->getBoxes());
  }
  // we should also define the levelSolver here as well

  PoissonOp thisLevelOp;
  thisLevelOp.define(a_grids, crseGridsPtr, a_dx, a_nRefCrse, 
		     a_domain, false);

  // set Projection physical BC's -- level projections use 
  // inhomogeneous form of BCs
  DomainGhostBC* physBCPtr = m_physBCPtr->LevelPressureBC();
  thisLevelOp.setDomainGhostBC(*physBCPtr);
  delete physBCPtr;

  m_levelSolver.define(a_grids, crseGridsPtr, a_domain, a_dx,
		       a_nRefCrse, &thisLevelOp);
  
  if (s_verbosity >= 3) {
    m_levelSolver.setVerbose(true);
  } else {
    m_levelSolver.setVerbose(false);
  }

  m_isInitialized = false;

}


// -------------------------------------------------------
void 
CCProjector::init(const CCProjector& a_oldProj)
{

  // this function performs a partial initialization 
  // of this object using the oldProj...

  // at the current state of this algorithm, doesn't
  // seem like this should do anything (everything will
  // need to be re-initialized anyway)

}

void 
CCProjector::variableSetUp()
{
  // first set up parm parse object
  ParmParse pp("projection");
  
  int tempBool;

  tempBool = (int) m_doSyncProjection;
  pp.query("doSyncProjection",tempBool);
  m_doSyncProjection = (tempBool == 1);

  tempBool = (int) m_applySyncCorrection;
  pp.query("applySyncCorrection", tempBool);
  m_applySyncCorrection = (tempBool == 1);

  tempBool = (int) s_applyVDCorrection;
  pp.query("applyFreestreamCorrection", tempBool);
  s_applyVDCorrection = (tempBool == 1);

  tempBool = (int) m_doQuadInterp;
  pp.query("doQuadInterp", tempBool);
  m_doQuadInterp = (tempBool == 1);

  pp.query("eta", m_etaLambda);

  // set flag so we don't have to call this function again.
  pp_init = true;

}
  

void
CCProjector::setCrseProj(CCProjector* a_crseProjPtr, int a_nRefCrse)
{
  if (m_crseProjPtr != NULL) {
    // if a coarse level is defined, should never undefine it
    assert (a_crseProjPtr != NULL);
    // nRefCrse should never change either.
    assert (m_nRefCrse == a_nRefCrse);
    // check to see if coarse-grid info will need to be redone
    if (!(m_crseProjPtr->getBoxes()==a_crseProjPtr->getBoxes())) {
      // new coarse boxes -- need to redefine levelSolver
      m_levelSolver.clearMemory();
      m_levelSolver.setDefaultValues();

      PoissonOp thisLevelOp;
      thisLevelOp.define(getBoxes(), &(a_crseProjPtr->getBoxes()), 
			 m_dx, m_nRefCrse, m_domain, false);
      // level projections will use inhomogeneous form of BCs
      DomainGhostBC* physBCPtr = m_physBCPtr->LevelPressureBC();
      thisLevelOp.setDomainGhostBC(*physBCPtr);
      delete physBCPtr;

      const DisjointBoxLayout& crseGrids = a_crseProjPtr->getBoxes();
      m_levelSolver.define(getBoxes(), &crseGrids, m_domain, m_dx,
			   a_nRefCrse, &thisLevelOp);

      if (s_verbosity >= 3) {
        m_levelSolver.setVerbose(true);
      } else {
        m_levelSolver.setVerbose(false);
      }

    } else {
      // if old and new coarse grids are the same, no need to 
      // do this
    }
  } else {
    // if old crse level didn't exist and new one does, then need
    // to initialize levelsolver
    if (a_crseProjPtr != NULL) {
      m_nRefCrse = a_nRefCrse;
      PoissonOp thisLevelOp;
      
      const DisjointBoxLayout& crseGrids = a_crseProjPtr->getBoxes();
      thisLevelOp.define(getBoxes(), &crseGrids, m_dx, m_nRefCrse,
			 m_domain, false);
      DomainGhostBC* physBCPtr = m_physBCPtr->LevelPressureBC();
      thisLevelOp.setDomainGhostBC(*physBCPtr);
      delete physBCPtr;

      m_levelSolver.define(getBoxes(), &crseGrids, m_domain, m_dx,
			   a_nRefCrse, &thisLevelOp);
      if (s_verbosity >= 3) {
        m_levelSolver.setVerbose(true);
      } else {
        m_levelSolver.setVerbose(false);
      }

    }

  }

  // don't forget to actually _set_ the pointer!
  m_crseProjPtr = a_crseProjPtr;
}

// -----------------------------------------------------------
void
CCProjector::setFineProj(CCProjector* a_fineProjPtr)
{
  // this is much simpler than resetting crseProj, since
  // there is no need to rebuild levelsolvers or anything like
  // that
  m_fineProjPtr = a_fineProjPtr;
}


#ifdef HDF5
// --------------------------------------------------------------
void 
CCProjector::writeCheckpointHeader(HDF5Handle& a_handle) const
{

  // i don't think anything needs to be done here -- just 
  // included for completeness...  
  
}

// --------------------------------------------------------------
void
CCProjector::writeCheckpointLevel(HDF5Handle& a_handle) const 
{
  
  char level_str[20];
  sprintf (level_str, "%d", m_level);
  const std::string label = std::string("level_") + level_str 
    + std::string("/projection");

  a_handle.setGroup(label);

  HDF5HeaderData header;
  
  header.m_real["dx"] = m_dx;
  header.m_box["domain"] = m_domain.domainBox();
  header.m_int["n_ref_crse"] = m_nRefCrse;
  if (m_finest_level) 
    header.m_int["finest_level"] = 1;
  else
    header.m_int["finest_level"] = 0;

  // as in AMREuler, don't write static data (comes from ParmParse)

  header.writeToFile(a_handle);

  // now dump out this level's data
  write (a_handle, m_Pi.boxLayout());
  // also want to dump out ghost-cell vales
  write (a_handle, m_Pi, "pi", m_Pi.ghostVect());
  write (a_handle, m_phi, "phi", m_phi.ghostVect());
  write (a_handle, m_eSync, "e_sync", m_eSync.ghostVect());
  write (a_handle, m_eLambda, "e_lambda", m_eLambda.ghostVect());
  write (a_handle, m_grad_eLambda, "grad_e_lambda", m_grad_eLambda.ghostVect());
  
}


// --------------------------------------------------------------
void
CCProjector::readCheckpointHeader(HDF5Handle& a_handle)
{
  // i don't think this actually does anything...
}

// --------------------------------------------------------------
void
CCProjector::readCheckpointLevel(HDF5Handle& a_handle)
{
 
  char level_str[20];
  sprintf (level_str, "%d", m_level);
  const std::string label = std::string("level_") + level_str 
    + std::string("/projection");

  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3) {
    pout () << "hdf5 header data: " << endl;
    pout () << header << endl;
  }

  // read dx
  if (header.m_real.find("dx") == header.m_real.end()) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain dx");
  }
  if (m_dx != header.m_real["dx"]) {
    MayDay::Error("CCProjector::readCheckpointLevel: dx in checkfile does not match that in solver");
  }
    
    
  // read domain
  if (header.m_box.find("domain") == header.m_box.end()) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain domain");
    if (m_domain.domainBox() != header.m_box["domain"]) {
      MayDay::Error("CCProjector::readCheckpointLevel: domain in checkfile does not match that in solver!");
    }
  }


  // read nRefCrse
  if (header.m_int.find("n_ref_crse") == header.m_int.end()) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain n_ref_crse");
    if (m_nRefCrse != header.m_int["n_ref_crse"]) {
      MayDay::Error("CCProjector::readCheckpointLevel: n_ref_crse in checkfile does not match that in solver!");
    }
  }

  // read nRefCrse
  if (header.m_int.find("finest_level") == header.m_int.end()) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain finest_level");
  }
  m_finest_level = (header.m_int["finest_level"] == 1);
  

  // read grids
  Vector<Box> grids;
  const int grid_status = read(a_handle, grids);
  if (grid_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain a Vector<Box>");
  }
  // can't really check for equality, but at least check for same size...
  if (grids.size() != m_Pi.getBoxes().size()) {
    MayDay::Error("CCProjector::readCheckpointLevel: grids in checkfile different size than in solver!");
  }

  const DisjointBoxLayout& levelGrids = m_Pi.getBoxes();
  const int piData_status = read<FArrayBox> (a_handle, m_Pi, "pi", levelGrids);
  
  if (piData_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain pi data");
  }

  const int phiData_status = read<FArrayBox> (a_handle, m_phi, "phi", levelGrids);
  
  if (phiData_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain phi data");
  }


  const int eSyncData_status = read<FArrayBox> (a_handle, m_eSync, "e_sync", levelGrids);
  
  if (eSyncData_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain e_sync data");
  }


  const int eLambdaData_status = read<FArrayBox> (a_handle, m_eLambda, "e_lambda", levelGrids); 
  
  if (eLambdaData_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain eLambda data");
  }


  const int grad_eLambdaData_status = read<FluxBox> (a_handle, m_grad_eLambda, "grad_e_lambda", levelGrids);
  
  if (grad_eLambdaData_status != 0) {
    MayDay::Error("CCProjector::readCheckpointLevel: file does not contain grad_e_lambda data");
  }
  

  // now need to set boundary conditions on all these things
  postRestart();

}

#endif

// --------------------------------------------------------------
// note that this function assumes that all BC's on phi have been set
void 
CCProjector::gradPhi(LevelData<FArrayBox>& a_gradPhi, int a_dir) const
{
  // first compute edge-centering of gradPhi
  DataIterator dit = a_gradPhi.dataIterator();
  dit.reset();

  int edgeDir = -1;
  const Box& edgeBox = a_gradPhi[dit()].box();
  for (int dir=0; dir<SpaceDim; dir++) {
    if (edgeBox.type(dir) == IndexType::NODE) {
      assert (edgeDir == -1);
      edgeDir = dir;
    }
  }
  assert(edgeDir != -1);


  const DisjointBoxLayout& grids = getBoxes();

  for (dit.reset(); dit.ok(); ++dit) {
    // call gradient subroutine directly, since this is one-directional
    
    // first need to set up corner boxes for internal use
    // corner boxes identify which cells need special
    // stencil for transverse derivatives
    // "bottom", "top", "left", and "right" are relative
    // to edgeDir=0, dir=1
    Box BLcorner(edgeBox);
    Box BRcorner(edgeBox);
    Box ULcorner(edgeBox);
    Box URcorner(edgeBox);
    
    // make sure that i check these boxes
    assert(false);
    
    // now loop over directions and set box edge extents 
    for (int tempDir=0; tempDir<SpaceDim; ++tempDir) {
      if ((tempDir==edgeDir)) {	    
	BLcorner.setBig(tempDir,edgeBox.smallEnd(tempDir));
	ULcorner.setBig(tempDir,edgeBox.smallEnd(tempDir));
	BRcorner.setSmall(tempDir,edgeBox.bigEnd(tempDir));
	URcorner.setSmall(tempDir,edgeBox.bigEnd(tempDir));
      } else if (tempDir == a_dir) {
	BLcorner.setBig(tempDir,edgeBox.smallEnd(tempDir));
	BRcorner.setBig(tempDir,edgeBox.smallEnd(tempDir));
	ULcorner.setSmall(tempDir,edgeBox.bigEnd(tempDir));
	URcorner.setSmall(tempDir,edgeBox.bigEnd(tempDir));
      }
    }
    
    FORT_MACGRAD(CHF_FRA1(a_gradPhi[dit()],0),
		 CHF_CONST_FRA1(m_phi[dit()],0),
		 CHF_BOX(grids[dit()]),
		 CHF_CONST_REAL(m_dx),
		 CHF_INT(a_dir),
		 CHF_INT(edgeDir),
		 CHF_BOX(BLcorner),
		 CHF_BOX(BRcorner),
		 CHF_BOX(ULcorner),
		 CHF_BOX(URcorner));
  } // end loop over grids
}


// --------------------------------------------------------------
// note that this function assumes that all BC's on phi have been set
void
CCProjector::gradPhi(LevelData<FluxBox>& a_gradPhi) const
{
  levelGradientMAC(a_gradPhi, m_phi, m_dx);
}


// --------------------------------------------------------------
// note that this function assumes that all BC's on Pi have been set
void
CCProjector::gradPi(LevelData<FArrayBox>& a_gradPi, int a_dir) const
{
  DataIterator dit = a_gradPi.dataIterator();
  
  for (dit.reset(); dit.ok(); ++dit) {
    // call FORTRAN subroutine directly...
    FORT_GRADCC(CHF_FRA1(a_gradPi[dit()],0),
		CHF_CONST_FRA1(m_phi[dit()],0),
		CHF_BOX(getBoxes()[dit()]),
		CHF_CONST_REAL(m_dx),
		CHF_INT(a_dir));
  }

}



// --------------------------------------------------------------
// note that this function assumes that all BC's on Pi have been set
void
CCProjector::gradPi(LevelData<FArrayBox>& a_gradPi) const
{

  levelGradientCC(a_gradPi, m_Pi, m_dx);

}


// -------------------------------------------------------------
// note that this function assumes all BC's on eSync have been set
// also only returns valid grad(eSync) in valid regions of grids
void
CCProjector::grad_eSync(LevelData<FArrayBox>& a_grad_eSync) const
{


  if (m_fineProjPtr != NULL) {
    const LevelData<FArrayBox>& fine_eSync = m_fineProjPtr->eSync();
    int nRefFine = m_fineProjPtr->nRefCrse();
    
    compGradientCC(a_grad_eSync, m_eSync,  &fine_eSync,
		   m_dx, nRefFine, m_domain);
  } else {
    // no finer level -- can do level-operator gradient
    levelGradientCC(a_grad_eSync, m_eSync, m_dx);
  }

}

// --------------------------------------------------------------
// note that this function assumes all BC's on eSync have been set
// also only returns valid grad(eSync) in valid regions of grids
void
CCProjector::grad_eSync(LevelData<FArrayBox>& a_grad_eSync, int a_dir) const
{
  // this is not yet implemented (not sure if i'll really need it!)
  MayDay::Warning("directional grad_eSync not yet implemented!");
}
      


// --------------------------------------------------------------
void
CCProjector::levelMacProject(LevelData<FluxBox>& a_uEdge,
			     Real a_oldTime, Real a_dt) 
{
  // MAC rhs should have no ghost values
  LevelData<FArrayBox> MacRHS(getBoxes(),1);

  levelDivergenceMAC(MacRHS, a_uEdge, m_dx);

  LevelData<FArrayBox>* crseBCPtr = NULL;
  Real CFscale = 0.5*a_dt;

  if (m_crseProjPtr != NULL) {
    // coarse-fine BC is 0.5*dt*(coarse Pi)
    const DisjointBoxLayout crseGrids = m_crseProjPtr->getBoxes();
    crseBCPtr = new LevelData<FArrayBox>(crseGrids,1);
    LevelData<FArrayBox>& crseBC = *crseBCPtr;

    Interval comps(0,0);
    const LevelData<FArrayBox>& crsePi = m_crseProjPtr->Pi();
    crsePi.copyTo(comps,crseBC,comps);

    DataIterator dit = crseBC.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      crseBC[dit()] *= CFscale;
    }
  }

  // now solve for phi
  m_levelSolver.levelSolve(m_phi, crseBCPtr, MacRHS);
  
  // now correct

  // first re-apply physical and copy BC's
  Interval phiComps(0,0);
  m_phi.exchange(phiComps);
  DomainGhostBC* physBCPtr = m_physBCPtr->gradMacPressureBC();

  DataIterator dit = m_phi.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    physBCPtr->applyInhomogeneousBCs(m_phi[dit()], m_domain, m_dx);
  }
  
  applyMacCorrection(a_uEdge, CFscale);

  // debugging check
  levelDivergenceMAC(MacRHS, a_uEdge, m_dx);  


  delete physBCPtr;

  if (crseBCPtr != NULL) {
    delete crseBCPtr;
    crseBCPtr = NULL;
  }
  // that should be it!
}


// --------------------------------------------------------------
void 
CCProjector::applyMacCorrection(LevelData<FluxBox>& a_uEdge,
				Real CFscale) 
{
  // this function computes a_uEdge - grad(phi) 
  // (assumes all BC's already set)

  const DisjointBoxLayout& levelGrids = a_uEdge.getBoxes();
  LevelData<FluxBox> gradPhi(levelGrids,1);

  LevelData<FArrayBox>* crseBCDataPtr=NULL;
  // should C/F BC's already be set? assume not and reset them here
  if (m_crseProjPtr != NULL) {

    const DisjointBoxLayout crseGrids = m_crseProjPtr->getBoxes();
    crseBCDataPtr = new LevelData<FArrayBox>(crseGrids,1);
    LevelData<FArrayBox>& crseBCData = *crseBCDataPtr;
    
    Interval comps(0,0);
    m_crseProjPtr->Pi().copyTo(comps,crseBCData,comps);
    
    // now scale the coarse-level BC
    DataIterator ditCrse = crseBCData.dataIterator();
    
    for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) {
      crseBCData[ditCrse()] *= CFscale;
    }
      
  }
  

  // compute gradient
  levelGradientMAC(gradPhi,m_phi, crseBCDataPtr, m_dx, m_cfInterp);
  
  // now rescale and subtract from uEdge
  DataIterator dit = a_uEdge.dataIterator();
  
  // assumes that uEdge and phi can use same dataIterator
  for (dit.reset(); dit.ok(); ++dit) {
    FluxBox& thisGrad = gradPhi[dit()];
    FluxBox& thisEdgeVel = a_uEdge[dit()];
    
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& thisGradDir = thisGrad[dir];
      FArrayBox& thisVelDir = thisEdgeVel[dir];

      thisVelDir -= thisGradDir;
    }
  }

  if (crseBCDataPtr!=NULL) {
    delete crseBCDataPtr;
    crseBCDataPtr = NULL;
  }

}

// ---------------------------------------------------------------
void 
CCProjector::LevelProject(LevelData<FArrayBox>& a_velocity,
			  LevelData<FArrayBox>* a_crseVelPtr,
			  const Real a_newTime, const Real a_dt)
{
  LevelData<FArrayBox>* crseDataPtr=NULL;
  LevelData<FArrayBox> levelProjRHS(getBoxes(),1);
  

  // just to be safe.  proper place for this may be outside this function
  Interval velComps(0,SpaceDim-1);
  a_velocity.exchange(velComps);

  if (m_level > 0) {
    if (doQuadInterp()) {
      assert(a_crseVelPtr != NULL);
      assert(m_crseProjPtr != NULL);
      
      // will need to add dt*gradPi from crse vel.
      const DisjointBoxLayout& crseLevelGrids = m_crseProjPtr->getBoxes();
      crseDataPtr= new LevelData<FArrayBox>(crseLevelGrids, SpaceDim);
      LevelData<FArrayBox>& crseData = *crseDataPtr;
      m_crseProjPtr->gradPi(crseData);
      
      // now multiply by dt and add crseVel
      DataIterator ditCrse = crseData.dataIterator();
      LevelData<FArrayBox>& crseVel = *a_crseVelPtr;

      // inherent assumption that crseVel and crseGradPi share
      // the same layout here
      for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) {
	FArrayBox& thisCrseVel = crseVel[ditCrse()];
	FArrayBox& thisCrseGradPi = crseData[ditCrse()];
	
	// note that we multiply by dt on _this_ level
	thisCrseGradPi *= a_dt;
	
	// now add crseVel
	thisCrseGradPi.plus(thisCrseVel);
      }
      // now copy crseBC values to crseBCData 
      //Interval comps(0,SpaceDim-1);
      //const DisjointBoxLayout& crseGrids = m_crseProjPtr->getBoxes();
      //crseDataPtr->define(crseGrids,SpaceDim);
      //crseGradPi.copyTo(comps, *crseDataPtr,comps);
      

    } else {
      // alternative is not implemented at this point
      MayDay::Error("non-quadratic interpolation BC's not implemented");
    }
  } // end if coarser level exists

  QuadCFInterp velCFInterp;
  if (doQuadInterp() && m_crseProjPtr != NULL) {
    // define two-component CF-interp object for velocities
    velCFInterp.define(a_velocity.getBoxes(), 
		       &(crseDataPtr->getBoxes()),
		       m_dx, m_nRefCrse, SpaceDim, m_domain);
  }

  // now compute RHS for projection
  levelDivergenceCC(levelProjRHS, a_velocity, crseDataPtr,
		    m_dx, doQuadInterp(), velCFInterp);

  // for proper scaling of pi, divide this by dt
  DataIterator dit = levelProjRHS.dataIterator();
  Real dtScale = 1.0/a_dt;
  for (dit.reset(); dit.ok(); ++dit) {
    levelProjRHS[dit()] *= dtScale;
  }

  // set up coarse BC's for solve, then solve
  const LevelData<FArrayBox>* crsePiPtr = NULL;
  if (m_crseProjPtr != NULL) crsePiPtr = &(m_crseProjPtr->Pi());
  
  // now solve for pi
  m_levelSolver.levelSolve(m_Pi, crsePiPtr, levelProjRHS);  
    
  // apply appropriate physical BC's
  DomainGhostBC* piBCPtr = m_physBCPtr->gradPiBC();
  // loop over grids...
  for (dit.reset(); dit.ok(); ++dit) {
    piBCPtr->applyInhomogeneousBCs(m_Pi[dit()], m_domain, m_dx);
  }
  delete piBCPtr;

  if (m_crseProjPtr != NULL) {
    // reapply coarse-fine BC's here if necessary
    m_cfInterp.coarseFineInterp(m_Pi, *crsePiPtr);
    // also clean up after ourselves!
    delete crseDataPtr;
    crseDataPtr = NULL;
     
  }

  Interval piComps(0,0);
  m_Pi.exchange(piComps);
  
  dtScale = a_dt;
  correctCCVelocities(a_velocity, dtScale);

  // check resulting velocity field
  // to do this, need to reset physical BC's

  basicVelBC* velBCPtr = m_physBCPtr->uStarBC(false);

  velBCPtr->applyBCs(a_velocity,m_domain, m_dx);
  
  levelDivergenceCC(levelProjRHS, a_velocity, a_crseVelPtr, 
		    m_dx, doQuadInterp(), velCFInterp);

  delete velBCPtr;
  
}
  
// ---------------------------------------------------------------
void
CCProjector::correctCCVelocities(LevelData<FArrayBox>& a_velocity,
				 const Real scale) const
{

  const DisjointBoxLayout& grids = getBoxes();
  LevelData<FArrayBox> gradPi(grids, SpaceDim);

  // assume that all relevant BC's have already been set
  levelGradientCC(gradPi, m_Pi, m_dx);

  // vel = vel - scale*gradPi

  DataIterator dit = gradPi.dataIterator();
  
  for (dit.reset(); dit.ok(); ++dit) {
    FArrayBox& thisGradPi = gradPi[dit()];
    FArrayBox& thisVel = a_velocity[dit()];
    
    thisGradPi *= scale;
    thisVel -= thisGradPi;
  }

}




// ---------------------------------------------------------------
void 
CCProjector::doSyncOperations(Vector<LevelData<FArrayBox>* >& a_velocity,
			      Vector<LevelData<FArrayBox>* >& a_lambda,
			      const Real a_newTime, const Real a_dtLevel)
{

  AMRSolver* bigSolver = new AMRSolver;

  defineSolver(*bigSolver, a_velocity, false);

  // now call sync projection
  doSyncProjection(a_velocity, a_newTime, a_dtLevel, *bigSolver);

  delete bigSolver;

  bigSolver = new AMRSolver;
  defineSolver(*bigSolver, a_lambda, true);

  // now do freestream preservation solve
  computeVDCorrection(a_lambda, a_newTime, a_dtLevel, *bigSolver);

  delete bigSolver;
}
		      
  
// ---------------------------------------------------------------
void
CCProjector::doPostRegridOps(Vector<LevelData<FArrayBox>* >& a_velocity,
			     Vector<LevelData<FArrayBox>* >& a_lambda,
			     const Real a_dt, const Real a_time)
{

  AMRSolver* bigSolverPtr = new AMRSolver;


#if 0
  defineSolver(*bigSolverPtr, a_velocity,false);

  // for post-regridding, CF BC should be homogeneous
  bool homogeneousCFBC = true;

  // now call initial velocity projection
  // this isn't really right -- really want to just project
  // new levels, not unchanged level as well.
  // however, making two AMRSolvers might be expensive...)
  initialVelocityProject(a_velocity, *bigSolverPtr,homogeneousCFBC);

  delete bigSolverPtr;
  bigSolverPtr = new AMRSolver;
#endif

  defineSolver(*bigSolverPtr, a_lambda, false);
  
  // for inviscid flow, only do this
  // now do freestream preservation solve
  computeVDCorrection(a_lambda, a_time, a_dt, *bigSolverPtr);
  
  delete bigSolverPtr;
}


// ---------------------------------------------------------------
void 
CCProjector::doSyncProjection(Vector<LevelData<FArrayBox>* >& a_velocity,
			      const Real a_newTime, const Real a_dtSync,
			      AMRSolver& a_solver) 
{

  if (doSyncProjection()) {
    assert(a_solver.isDefined());
    
    // set up temp storage
    int vectorSize = a_velocity.size();
    
    Vector<LevelData<FArrayBox>* > syncRHS(vectorSize,NULL);
    Vector<LevelData<FArrayBox>* > syncCorr(vectorSize,NULL);
    
    // initially 

    CCProjector* levelProjPtr = this;
    while (!levelProjPtr->isFinestLevel()) {
      levelProjPtr = levelProjPtr->fineProjPtr();
    }
    
    int finestLevel = levelProjPtr->getLevel();
    
    // reset projection ptr to this level
    levelProjPtr = this;
    
    // this is a bogus leveldata pointer for the composite divergences
    LevelData<FArrayBox>* crseVelPtr=NULL;
    LevelData<FArrayBox>* fineVelPtr = NULL;
    
    // loop over levels to allocate storage and compute RHS
    for (int lev = m_level; lev<= finestLevel; lev++) {
      const DisjointBoxLayout& levelGrids = a_velocity[lev]->getBoxes();
      
      syncRHS[lev]= new LevelData<FArrayBox>(levelGrids,1);
      syncCorr[lev] = &(levelProjPtr->eSync());
      
      
      Real dx = levelProjPtr->dx();
      const ProblemDomain& levelDomain = levelProjPtr->dProblem();
      int nRefCrse = levelProjPtr->nRefCrse();
      int nRefFine = -1;
      
      if (lev > 0) crseVelPtr = a_velocity[lev-1];
      if (lev < finestLevel) {
	fineVelPtr = a_velocity[lev+1];
	nRefFine = levelProjPtr->fineProjPtr()->nRefCrse();
      } else {
	fineVelPtr = NULL;
      }

      // do exchange here
      a_velocity[lev]->exchange(a_velocity[lev]->interval());
      
      // now compute divergence
      compDivergenceCC(*syncRHS[lev],*a_velocity[lev],crseVelPtr,
		       fineVelPtr, dx, nRefCrse, nRefFine, levelDomain, 
		       doQuadInterp());

      // increment projection level
      levelProjPtr = levelProjPtr->fineProjPtr();
      
    } // end loop over levels for computation of RHS
    
    // take care of C/F Boundary condition for lBase
    if (m_level > 0) {
      syncCorr[m_level-1] = &(m_crseProjPtr->eSync());
      // need to rescale crse-level BC 
      LevelData<FArrayBox>& crseSyncCorr = *(syncCorr[m_level-1]);
      DataIterator ditCrse = crseSyncCorr.dataIterator();
      for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) {
	crseSyncCorr[ditCrse()] *= a_dtSync;
      }
    }
    
    
    // debugging tool -- want sum(rhs) = 0
    Vector<int> nRefFineVect(syncRHS.size());
    levelProjPtr = this;
    for (int lev=m_level; lev<=finestLevel; lev++) {
      if (lev < finestLevel) {
	nRefFineVect[lev] = levelProjPtr->fineProjPtr()->nRefCrse();
      } else {
	nRefFineVect[lev] = 1;
      }
      levelProjPtr = levelProjPtr->fineProjPtr();
    }


    Interval sumComps(0,0);
    Real sumRHS;
    sumRHS =  computeSum(syncRHS, nRefFineVect, m_dx, sumComps, m_level);
    pout() << "Sum(RHS) for sync solve = " 
           << setiosflags(ios::scientific) << sumRHS << endl;

      
    
    // now solve   
    a_solver.solveAMR(syncCorr, syncRHS);

#if 0
#ifdef HDF5
  // dump out projection RHS
  Vector<string> rhsNames(1);
  rhsNames[0] = "syncRHS";

  Vector<DisjointBoxLayout> levelGrids(finestLevel+1);
  for (int lev=0; lev<=finestLevel; lev++) {
    levelGrids[lev] = syncCorr[lev]->getBoxes();
  }

  WriteAMRHierarchyHDF5("syncProjRHS.hdf5",
			levelGrids, 
			syncRHS,
			rhsNames,
			m_domain,
			m_dx, 1.0, 1.0, 
			nRefFineVect,
			levelGrids.size());


  Vector<string> corrNames(1);
  corrNames[0] = "syncCorr";

  WriteAMRHierarchyHDF5("syncCorr.hdf5",
			levelGrids, 
			syncCorr,
			corrNames,
			m_domain,
			m_dx, 1.0, 1.0, 
			nRefFineVect,
			levelGrids.size());
#endif

#endif

#if 0 // this is actually done in the composite gradient, isn't it?
    // just in case, re-apply coarse-fine BC's to syncCorr
    levelProjPtr = this;
    for (int lev=m_level; lev<= finestLevel; lev++) {
      if (lev < 0) {
	QuadCFInterp& levelCFInterp = levelProjPtr->quadCFInterpolator();
	levelCFInterp.coarseFineInterp(syncCorr[lev], syncCorr[lev-1]);
      }
      levelProjPtr = levelProjPtr->fineProjPtr();
    }
    
#endif 

    // apply physical boundary conditions before
    // taking gradients
    Interval corrComps(0,0);
    levelProjPtr = this;

    DomainGhostBC* eSyncBCPtr = m_physBCPtr->gradESyncBC();
    for (int lev=m_level; lev<=finestLevel; lev++) {

      // loop over grids
      LevelData<FArrayBox>& level_eSync = *syncCorr[lev];
      const ProblemDomain& levelDomain = levelProjPtr->dProblem();
      Real levelDx = levelProjPtr->dx();
      DataIterator ditLev = level_eSync.dataIterator();

      for (ditLev.reset(); ditLev.ok(); ++ditLev) {
	eSyncBCPtr->applyInhomogeneousBCs(level_eSync[ditLev()], levelDomain,
					  levelDx);
      }

      level_eSync.exchange(corrComps);
      levelProjPtr = levelProjPtr->fineProjPtr();
    }

    delete eSyncBCPtr;
    
    // now apply sync correction
    Real scale = 1.0;
    LevelData<FArrayBox>* crseCorrPtr = NULL;
    if (m_level > 0) crseCorrPtr = syncCorr[m_level-1];
    
    applySyncCorrection(a_velocity, scale, crseCorrPtr);


    // cleaning up after we're done -- 
    // now need to rescale eSync to look like other pressures
    // start with level-1 because we rescaled it for the BC
    // also need to delete syncRHS

    Real corrScale = 1.0/a_dtSync;
    for (int lev=m_level-1; lev<=finestLevel; lev++) {
      if (lev >=0) {
	
	LevelData<FArrayBox>& levelCorr = *syncCorr[lev];
	DataIterator dit = levelCorr.dataIterator();
	for (dit.reset(); dit.ok(); ++dit) {
	  levelCorr[dit()] *= corrScale;
	}
	levelCorr.exchange(corrComps);
      } // end if level > 0
    } // end loop over levels for rescaling eSync
    

    // loop over levels to delete syncRHS
    for (int lev= m_level; lev<=finestLevel; lev++) {
      delete syncRHS[lev];
      syncRHS[lev] = NULL;
    }



  } // end if do sync in the first place

}
 
// ---------------------------------------------------------------
void
CCProjector::applySyncCorrection(Vector<LevelData<FArrayBox>* >& a_velocities,
				 const Real a_scale,
				 LevelData<FArrayBox>* crseCorrPtr)
{
  
  if (m_applySyncCorrection) {
    
    const DisjointBoxLayout& levelGrids = getBoxes();
    
    LevelData<FArrayBox> levelGradCorr(levelGrids, SpaceDim);
    
    LevelData<FArrayBox>* fineCorrPtr = NULL;
    int nRefFine = -1;
    
    if (!isFinestLevel()) {
      fineCorrPtr = &(m_fineProjPtr->eSync());
      nRefFine = m_fineProjPtr->nRefCrse();
    }
    
    compGradientCC(levelGradCorr, m_eSync, crseCorrPtr, fineCorrPtr, 
		   m_dx, nRefFine, m_domain, m_cfInterp);
    
    // now loop over grids and correct
    LevelData<FArrayBox>& levelVel = *a_velocities[m_level];
    
    DataIterator dit = levelVel.dataIterator();
    
    for (dit.reset(); dit.ok(); ++dit) {
      levelGradCorr[dit()] *= a_scale;
      levelVel[dit()] -= levelGradCorr[dit()];
    }
    
    
    // now recursively call finer level
    if (!isFinestLevel()) {
      m_fineProjPtr->applySyncCorrection(a_velocities, a_scale,
					 &m_eSync);

      // average down finer level to this level
      const DisjointBoxLayout& fineGrids = m_fineProjPtr->getBoxes();
      int nRefFine = m_fineProjPtr->nRefCrse();
      CoarseAverage avgDownThingy(fineGrids, SpaceDim, nRefFine);
      
      avgDownThingy.averageToCoarse(*a_velocities[m_level],
				    *a_velocities[m_level+1]);
      
    }
    
  }

}



// ---------------------------------------------------------------
void
CCProjector::computeVDCorrection(Vector<LevelData<FArrayBox>* >& a_lambda,
				 const Real a_newTime, const Real a_dtSync,
				 AMRSolver& a_solver) 
{
  
  // only do this if eta > 0
  if (m_etaLambda > 0) {
				 
    assert(a_solver.isDefined());
    
    // set up temp storage
    int vectorSize = a_lambda.size();
    
    Vector<LevelData<FArrayBox>* > VDCorr(vectorSize);
    
    CCProjector* levelProjPtr = this;
    while (!levelProjPtr->isFinestLevel()) {
      levelProjPtr = levelProjPtr->fineProjPtr();
    }
    
    int finestLevel = levelProjPtr->getLevel();
    
    // reset projection ptr to this level
    levelProjPtr = this;
    
    assert (a_dtSync != 0);
    Real lambdaMult = m_etaLambda/a_dtSync;

    for (int lev = m_level; lev<=finestLevel; lev++) {
      // stuff eLambda into VDCorr array
      VDCorr[lev] = &(levelProjPtr->eLambda());
      
      // use lambda array as RHS for solve
      LevelData<FArrayBox>& thisLambda = *(a_lambda[lev]);

      DataIterator dit = thisLambda.dataIterator();

      // RHS = (lambda-1)*eta/dtSync
      for (dit.reset(); dit.ok(); ++dit) {
	thisLambda[dit()] -= 1.0;
	thisLambda[dit()] *= lambdaMult;
	levelProjPtr->eLambda()[dit()].setVal(0.0);
      }

      // advance projection pointer to next finer level
      levelProjPtr = levelProjPtr->fineProjPtr();
    } // end loop over levels

    // if necessary, define coarse BC here
    if (m_level > 0) {
      VDCorr[m_level-1] = &(m_crseProjPtr->eLambda());
      // rescale crse-level BC
      DataIterator ditCrse = VDCorr[m_level-1]->dataIterator();
      for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) 
        {
          // try setting coarse-level BC to 0
          m_crseProjPtr->eLambda()[ditCrse()].setVal(0.0);
#if 0
          Real scale = m_dx/(m_crseProjPtr->m_dx);
          m_crseProjPtr->eLambda()[ditCrse()].mult(scale);
#endif

        }
    }

    // debugging tool -- want sum(rhs) = 0
    Vector<int> nRefFineVect(a_lambda.size());
    levelProjPtr = this;
    for (int lev=m_level; lev<=finestLevel; lev++) {
      if (lev < finestLevel) {
	nRefFineVect[lev] = levelProjPtr->fineProjPtr()->nRefCrse();
      } else {
	nRefFineVect[lev] = 1;
      }
      levelProjPtr = levelProjPtr->fineProjPtr();
    }


    Interval sumComps(0,0);
    Real sumRHS;
    sumRHS =  computeSum(a_lambda, nRefFineVect, m_dx, sumComps, m_level);
    pout() << "Sum(RHS) for VD solve = " 
           << setiosflags(ios::scientific) << sumRHS << endl;


    // now solve
    a_solver.solveAMR(VDCorr, a_lambda);

    // apply all bc's here

    DomainGhostBC* lambdaBCPtr = m_physBCPtr->gradELambdaBC();

    Interval corrComps(0,0);
    levelProjPtr = this;
    for (int lev=m_level; lev<=finestLevel; lev++) {
      LevelData<FArrayBox>& levelCorr = *(VDCorr[lev]);
      const ProblemDomain& levelDomain = levelProjPtr->dProblem();
      Real levelDx = levelProjPtr->dx();
      DataIterator ditLev = levelCorr.dataIterator();
      for (ditLev.reset(); ditLev.ok(); ++ditLev) {
	lambdaBCPtr->applyInhomogeneousBCs(levelCorr[ditLev()],
				       levelDomain, levelDx);
      }
      levelCorr.exchange(corrComps);
      levelProjPtr = levelProjPtr->fineProjPtr();
    }

    delete lambdaBCPtr;

    // compute correction field
    computeGrad_eLambda();
    // rescale correction field
    //rescaleGrad_eLambda(m_dx);

    lambdaMult = 1.0/lambdaMult;
    // now return lambda to previous state
    for (int lev = m_level; lev<=finestLevel; lev++) {
      LevelData<FArrayBox>& levelLambda = *(a_lambda[lev]);
      DataIterator dit = levelLambda.dataIterator();
      
      for (dit.reset(); dit.ok(); ++dit) {
	levelLambda[dit()] *= lambdaMult;
	levelLambda[dit()] += 1.0;
      }

    }

  }

}



    
// ---------------------------------------------------------------
void
CCProjector::computeGrad_eLambda()
{
  // do this recursively starting at the finest level

  if (!isFinestLevel()) {
    m_fineProjPtr->computeGrad_eLambda();
  }

  // now compute composite gradient on this level
  LevelData<FArrayBox>* crseBCPtr = NULL;
  
  if (m_crseProjPtr != NULL) crseBCPtr = &(m_crseProjPtr->eLambda());

  // recall that composite MAC gradient is the same as the level
  // gradient, since finer level is not considered to be part
  // of this level (take care of covered regions with avgDown)

  levelGradientMAC(m_grad_eLambda, m_eLambda, crseBCPtr, m_dx, 
		   m_cfInterp);


  // if s_applyVDCorrection is false, set to zero
  if (!s_applyVDCorrection)
    {
      DataIterator dit = m_grad_eLambda.dataIterator();
      for (dit.begin(); dit.ok(); ++dit)
        {
          m_grad_eLambda[dit()].setVal(0.0);
        }
    }


  // now average down finer level grad_eLambda if necessary
  if (!isFinestLevel()) {
    const DisjointBoxLayout& fineGrids = m_fineProjPtr->getBoxes();
    int nRefFine = m_fineProjPtr->nRefCrse();
    int nComp = 1;

    CoarseAverageEdge avgDownObject(fineGrids, nComp, nRefFine);

    const LevelData<FluxBox> & fineEdgeGrad = m_fineProjPtr->grad_eLambda();
    
    avgDownObject.averageToCoarse(m_grad_eLambda, fineEdgeGrad);
  }

  // do this as a diagnostic
  const DisjointBoxLayout& levelGrids = m_grad_eLambda.getBoxes();
  LevelData<FArrayBox> divUp(levelGrids, 1);

  levelDivergenceMAC(divUp, m_grad_eLambda, m_dx);


}
      
  
// ---------------------------------------------------------------
void
CCProjector::rescaleGrad_eLambda(Real a_dx_lbase)
{
  // do this recursively starting at the finest level

  if (!isFinestLevel()) {
    m_fineProjPtr->rescaleGrad_eLambda(a_dx_lbase);
  }

  // now rescale correction on this level.  we rescale correction
  // by (dt(lbase)/dt(this_level) = (dx(lbase)/m_dx for consistency
  Real scale = a_dx_lbase/m_dx;
  DataIterator dit = m_grad_eLambda.dataIterator();
  for (dit.begin(); dit.ok(); ++dit) 
    {
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          m_grad_eLambda[dit()][dir] *= scale;
        }
    }
}
		      
// ---------------------------------------------------------------
void
CCProjector::initialLevelProject(LevelData<FArrayBox>& a_velocity,
				 LevelData<FArrayBox>* a_crseVelPtr,
				 const Real a_newTime, const Real a_dt)
{
  
  LevelData<FArrayBox>* crseDataPtr=NULL;
  LevelData<FArrayBox> levelProjRHS(getBoxes(),1);

  // do coarse-fine BC
  if (m_level > 0) {
    if (doQuadInterp()) {
      assert(a_crseVelPtr != NULL);
      assert(m_crseProjPtr != NULL);
      
      // will need to add dt*gradPi to crse vel      
      const DisjointBoxLayout& crseLevelGrids = m_crseProjPtr->getBoxes();
      crseDataPtr = new LevelData<FArrayBox>(crseLevelGrids, SpaceDim);
      LevelData<FArrayBox>& crseData = *crseDataPtr;
      m_crseProjPtr->gradPi(crseData);

      // now multiply by dt and add crseVel
      DataIterator ditCrse = crseData.dataIterator();
      LevelData<FArrayBox>& crseVel = *a_crseVelPtr;
      
      // inherent assumption that crseVel & crseGradPi share same layout
      for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) {
	FArrayBox& thisCrseVel = crseVel[ditCrse()];
	FArrayBox& thisCrseGradPi = crseData[ditCrse()];

	// note that we multiply by this level's dt
	thisCrseGradPi *= a_dt;
	// now add crseVel
	thisCrseGradPi.plus(thisCrseVel);
      }

    } else {
      // alternative is not implemented at this point
      MayDay::Error("non-quadratic interpolation BC's not implemented");
    }
  } // end if coarser level exists

  // just to be safe...
  Interval velComps(0,SpaceDim-1);
  a_velocity.exchange(velComps);
  
  // now compute RHS for projection
  levelDivergenceCC(levelProjRHS, a_velocity, crseDataPtr,
		    m_dx, doQuadInterp(), m_cfInterp);
  
  // for proper scaling of pi, divide this by dt
  DataIterator dit = levelProjRHS.dataIterator();
  Real dtScale = 1.0/a_dt;

  for(dit.reset(); dit.ok(); ++dit) {
    levelProjRHS[dit()] *= dtScale;
  }

  // set up coarse BC's for solve, then solve
  const LevelData<FArrayBox>* crsePiPtr = NULL;
  if (m_crseProjPtr != NULL) crsePiPtr = &(m_crseProjPtr->Pi());
  
  // now solve for Pi
  m_levelSolver.levelSolve(m_Pi, crsePiPtr, levelProjRHS);

  if (m_crseProjPtr != NULL) {
    // reapply coarse-fine BC's here if necessary
    m_cfInterp.coarseFineInterp(m_Pi, *crsePiPtr);
  }


  // apply appropriate physical BC's
  DomainGhostBC* piBCPtr = m_physBCPtr->gradPiBC();
  // loop over grids...
  for (dit.reset(); dit.ok(); ++dit) {
    piBCPtr->applyInhomogeneousBCs(m_Pi[dit()], m_domain, m_dx);
  }
  
  delete piBCPtr;
  Interval piComps(0,0);
  m_Pi.exchange(piComps);

  dtScale = a_dt;
  correctCCVelocities(a_velocity,dtScale);

  // clean up storage
  if (crseDataPtr!=NULL) {
    delete crseDataPtr;
    crseDataPtr=NULL;
  }

}



// ---------------------------------------------------------------
void
CCProjector::initialSyncProjection(Vector<LevelData<FArrayBox>* >& a_vel,
				   const Real a_newTime, const Real a_dtSync,
				   AMRSolver& a_solver)
{

  // actually, in the current algorithm, i _think_ this is the same
  // as the normal sync projection
  doSyncProjection(a_vel, a_newTime, a_dtSync, a_solver);
}


// ---------------------------------------------------------------
void
CCProjector::doInitialSyncOperations(Vector<LevelData<FArrayBox>* >& a_vel,
				     Vector<LevelData<FArrayBox>* >& a_lambda,
				     const Real a_newTime, const Real a_dtSync)
{

  // now can define multilevel solver
  AMRSolver* bigSolverPtr = new AMRSolver;

  defineSolver(*bigSolverPtr, a_vel,false);
  
  // now call sync projection
  initialSyncProjection(a_vel, a_newTime, a_dtSync, *bigSolverPtr);

  
  delete bigSolverPtr;
  bigSolverPtr = new AMRSolver;
  defineSolver(*bigSolverPtr, a_lambda, true);

  // now do freestream preservation solve
  computeVDCorrection(a_lambda, a_newTime, a_dtSync, *bigSolverPtr);

  
}


// ---------------------------------------------------------------
void
CCProjector::initialVelocityProject(Vector<LevelData<FArrayBox>* >& a_vel,
				    bool a_homogeneousCFBC)
{

  // first need to define solver
  AMRSolver bigSolver;
  defineSolver(bigSolver,a_vel, false);

  initialVelocityProject(a_vel, bigSolver, a_homogeneousCFBC);
}


// ---------------------------------------------------------------
void 
CCProjector::initialVelocityProject(Vector<LevelData<FArrayBox>* >& a_vel,
				    AMRSolver& a_bigSolver,
				    bool a_homogeneousCFBC)
{
 
  // set up temp storage
  int vectorSize = a_vel.size();

  Vector<LevelData<FArrayBox>* > projRHS(vectorSize,NULL);
  Vector<LevelData<FArrayBox>* > projCorr(vectorSize,NULL);

  CCProjector* levelProjPtr = this;

  int finestLevel = vectorSize-1;


  // this is a bogus levelData pointer for the composite divergence
  LevelData<FArrayBox>* crseVelPtr = NULL;
  LevelData<FArrayBox>* fineVelPtr = NULL;

  // loop over levels to allocate storage and compute RHS
  for (int lev=m_level; lev <= finestLevel; lev++) {
    const DisjointBoxLayout& levelGrids = a_vel[lev]->getBoxes();
    
    projRHS[lev] = new LevelData<FArrayBox>(levelGrids,1);
    projCorr[lev] = &(levelProjPtr->eSync());

    Real dx = levelProjPtr->dx();
    const ProblemDomain& levelDomain = levelProjPtr->dProblem();
    int nRefCrse = levelProjPtr->nRefCrse();
    int nRefFine = -1;

    if (lev > 0) crseVelPtr = a_vel[lev-1];
    if (lev < finestLevel) {
      fineVelPtr = a_vel[lev+1];
      nRefFine = levelProjPtr->fineProjPtr()->nRefCrse();
    } else {
      fineVelPtr = NULL;
    }

    // just in case...
    Interval velComps(0,SpaceDim-1);
    a_vel[lev]->exchange(velComps);

    // now compute divergence
    compDivergenceCC(*projRHS[lev],*a_vel[lev],crseVelPtr,
		     fineVelPtr,dx,nRefCrse,nRefFine,levelDomain,
		     doQuadInterp());

    // increment levelProjPtr
    levelProjPtr= levelProjPtr->fineProjPtr();

  } // end loop over levels for computation of RHS

  // take care of C/F Boundary condition for lBase

  LevelData<FArrayBox>* crseBCPtr = NULL;
  if (m_level > 0) {
    assert(m_crseProjPtr != NULL);
    if (a_homogeneousCFBC) {
      // need to define coarse BC and set to 0
      const DisjointBoxLayout& crseGrids = m_crseProjPtr->getBoxes();

      crseBCPtr = new LevelData<FArrayBox>(crseGrids,1);
      
      // now set to 0
      LevelData<FArrayBox>& crseBC = *crseBCPtr;
      DataIterator ditCrse = crseBC.dataIterator();
      
      for (ditCrse.reset(); ditCrse.ok(); ++ditCrse) {
	crseBC[ditCrse()].setVal(0.0);
      }
      projCorr[m_level-1] = crseBCPtr;
    } else {
      // not sure if this should ever be called
      MayDay::Warning("inhomogeneous BC's on initialVelProject");
      projCorr[m_level-1] = &m_crseProjPtr->eSync();
    }

  } // end if level > 0


  Vector<DisjointBoxLayout> levelGrids(finestLevel+1);
  Vector<int> nRefFineVect(finestLevel+1);
  for (int lev=m_level; lev<=finestLevel; lev++) {
    levelGrids[lev] = projCorr[lev]->getBoxes();
    // hack this for now (assumes single nRef)
    if (crseProjPtr() != NULL) {
      nRefFineVect[lev] = m_nRefCrse; 
    } else if (!isFinestLevel()) {
      nRefFineVect[lev] = fineProjPtr()->nRefCrse();
    } else {
      // no other levels
      nRefFineVect[lev] = 1;
    }
  }



#if 0
#ifdef HDF5
  // dump out projection RHS
  Vector<string> rhsNames(1);
  rhsNames[0] = "syncRHS";

  WriteAMRHierarchyHDF5("initProjRHS.hdf5",
			levelGrids, 
			projRHS,
			rhsNames,
			m_domain,
			m_dx, 1.0, 1.0, 
			nRefFineVect,
			levelGrids.size());
#endif
#endif

  Interval sumComps(0,0);
  Real sumRHS;
  sumRHS =  computeSum(projRHS, nRefFineVect, m_dx, sumComps, m_level);
  pout() << "Sum(RHS) for IVP solve = " 
         << setiosflags(ios::scientific) << sumRHS << endl;


  // now solve!
  a_bigSolver.solveAMR(projCorr, projRHS);

#if 0
#ifdef HDF5
  // dump out projection RHS
  Vector<string> corrNames(1);
  corrNames[0] = "syncCorr";

  WriteAMRHierarchyHDF5("initProjCorr.hdf5",
			levelGrids, 
			projCorr,
			corrNames,
			m_domain,
			m_dx, 1.0, 1.0, 
			nRefFineVect,
			levelGrids.size());
#endif
#endif

  // apply physical boundary conditions before
  // taking gradients
  Interval corrComps(0,0);
  levelProjPtr = this;
  DomainGhostBC*  eSyncBCPtr = m_physBCPtr->gradESyncBC();
    
  for (int lev=m_level; lev<=finestLevel; lev++) {

    // loop over grids
    LevelData<FArrayBox>& levelCorr = *projCorr[lev];
    const ProblemDomain& levelDomain = levelProjPtr->dProblem();
    Real levelDx = levelProjPtr->dx();
    DataIterator ditLev = levelCorr.dataIterator();
    
    for (ditLev.reset(); ditLev.ok(); ++ditLev) {
      eSyncBCPtr->applyInhomogeneousBCs(levelCorr[ditLev()], levelDomain,
				    levelDx);
    }
    
    levelCorr.exchange(corrComps);

    levelProjPtr = levelProjPtr->fineProjPtr();
  }
  
  delete eSyncBCPtr;

  // now apply correction
  Real scale = 1.0;

  
  applySyncCorrection(a_vel, scale, crseBCPtr);

  // delete temp storage -- RHS and crse BC
  for (int lev=m_level; lev<=finestLevel; lev++) {
    delete projRHS[lev];
    projRHS[lev] = NULL;
  }

  if (crseBCPtr != NULL) {
    delete crseBCPtr;
    crseBCPtr = NULL;
  }
      
  


}



// ---------------------------------------------------------------
void
CCProjector::defineSolver(AMRSolver& a_solver, 
			  const Vector<LevelData<FArrayBox>* >& a_vel,
			  bool a_freestreamSolve)
{

  // assume that this level is lBase...
  int vectorSize = a_vel.size();


  // determine finest existing level
  CCProjector* levelProj = this;

  int finestLevel = vectorSize-1;
  int numLevels = finestLevel + 1;

  // quick sanity check
  //  assert(finestLevel < vectorSize);

  // put level grids togther into a vector
  Vector<DisjointBoxLayout> allGrids(numLevels);
  Vector<ProblemDomain> allDomains(numLevels);
  Vector<Real> allDx(numLevels);
  Vector<int> allRefRatios(numLevels);

  levelProj = this;
  
  for (int lev = m_level; lev<=finestLevel; lev++) {
    
    // for now, get grids from velocity arrays
    const DisjointBoxLayout& levelGrids = a_vel[lev]->getBoxes();
    allGrids[lev] = levelGrids;
    allDomains[lev] = levelProj->dProblem();
    allDx[lev] = levelProj->dx();
    
    // since nRef should be refinement ratio to next finer level,
    // advance the levelProj pointer to next finer level here.
    if (!levelProj->isFinestLevel()) {
      levelProj = levelProj->fineProjPtr();
      allRefRatios[lev] = levelProj->nRefCrse();
    }
  }

  // if coarser level exists. also need to include that for boundary
  // conditions

  //DisjointBoxLayout* crseGridsPtr = NULL;  //not used?  (ndk)
  if (m_level > 0) {
    assert(m_crseProjPtr != NULL);
    const DisjointBoxLayout& crseLevelGrids 
      = a_vel[m_level-1]->getBoxes();
    allGrids[m_level-1] = crseLevelGrids;
    //crseGridsPtr = &allGrids[m_level-1]; //not used?  (ndk)
    allDomains[m_level-1] = m_crseProjPtr->dProblem();
    allDx[m_level-1] = m_crseProjPtr->dx();
    allRefRatios[m_level-1] = nRefCrse();
  }

  PoissonOp bogusPoissonOp;
  
  
  //  bogusPoissonOp.define(getBoxes(), &allGrids[m_level], m_dx, m_nRefCrse,
  //		m_domain, false);

  DomainGhostBC* physBCPtr;
  if (a_freestreamSolve) {
    physBCPtr = m_physBCPtr->FreestreamCorrBC();
  } else {
    physBCPtr= m_physBCPtr->SyncProjBC();
  }
  bogusPoissonOp.setDomainGhostBC(*physBCPtr);
  delete physBCPtr;
  

  // now can define multilevel elliptic solver
  a_solver.define(allGrids, allDomains, allDx, allRefRatios, 
		  numLevels, m_level, &bogusPoissonOp);


  if (s_verbosity >= 3) {
    a_solver.setVerbose(true);
  } else {
    a_solver.setVerbose(false);
  }


}


// -------------------------------------------------------------
void
CCProjector::postRestart() 
{

  // set C/F BC's -- this depends on the fact that the coarser level
  // has already been read in!
  if (m_level > 0) {
    // need to do Pi 
    assert (m_cfInterp.isDefined());
    LevelData<FArrayBox>& crsePi = m_crseProjPtr->Pi();
    m_cfInterp.coarseFineInterp(m_Pi, crsePi);

  }

  Interval piComps = m_Pi.interval();
  m_Pi.exchange(piComps);

  // also need to do physical BC's
  DomainGhostBC* piBCPtr = m_physBCPtr->gradPiBC();

  // loop over grids
  DataIterator dit = m_Pi.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    piBCPtr->applyInhomogeneousBCs(m_Pi[dit()], m_domain, m_dx);
  }

  delete piBCPtr;
}
